"""Declares :class:`PlatformAgreementLanguageAdmin`."""
from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from iam.ext.django.models import PlatformAgreementLanguage


class PlatformAgreementLanguageAdmin(admin.ModelAdmin):
    pass


admin.site.register(PlatformAgreementLanguage, PlatformAgreementLanguageAdmin)



#pylint: skip-file
import django.test
import unimatrix.lib.test
from django.contrib.auth import authenticate
from django.contrib.auth import login

from iam.const import TEST_USER1_PHONENUMBER
from iam.const import TEST_USER1_ID
from iam.const import TEST_PASSWORD
from iam.domain import PhonenumberPrincipal
from iam.domain import NullCredential
from iam.domain import PasswordCredential
from iam.ext.django.lib.backends import LocalSubjectBackend

from .test_login_email_and_password import PasswordAuthenticationTestCase


@unimatrix.lib.test.integration
class PasswordAuthenticationTestCase(PasswordAuthenticationTestCase):
    authenticate_kwargs = {
        'request': None,
        'principal': PhonenumberPrincipal(TEST_USER1_PHONENUMBER),
        'credential': PasswordCredential(TEST_PASSWORD)
    }



"""Declares :class:`~iam.ext.django.models.Principal`."""
from django.db import models


class Principal(models.Model):
    """Generic object that maintains :term:`Principals`."""

    #: Maps domain object type identifiers to internal storage identifiers.
    storage_kinds = {
        'iam.unimatrixone.io/v1/FacebookAccountPrincipal': 'id.iam.unimatrixone.io/facebook',
        'iam.unimatrixone.io/v1/EmailPrincipal': 'id.iam.unimatrixone.io/email',
        'iam.unimatrixone.io/v1/GoogleAccountPrincipal': 'id.iam.unimatrixone.io/google',
        'iam.unimatrixone.io/v1/LinkedInAccountPrincipal': 'id.iam.unimatrixone.io/linkedin',
        'iam.unimatrixone.io/v1/NullPrincipal': 'id.iam.unimatrixone.io/null',
        'iam.unimatrixone.io/v1/PhonenumberPrincipal': 'id.iam.unimatrixone.io/phonenumber',
        'iam.unimatrixone.io/v1/TwitterAccountPrincipal': 'id.iam.unimatrixone.io/twitter',
    }

    #: A surrogate primary key for this :term:`Principal`.
    id = models.BigAutoField(
        primary_key=True,
        blank=False,
        null=False,
        db_column='id'
    )

    #: Specifies the :class:`~iam.ext.django.models.Realm` to which the
    #: :term:`Principal` belongs.
    realm = models.ForeignKey(
        'iam.Realm',
        on_delete=models.PROTECT,
        blank=False,
        null=False,
        db_column='realm_name'
    )

    #: Specifies the :term:`Subject` that is identified by the principal.
    subject = models.ForeignKey(
        'iam.Subject',
        on_delete=models.CASCADE,
        blank=False,
        null=False,
        related_name='principals',
        db_column='subject_id'
    )

    kind = models.CharField(
        max_length=64,
        blank=False,
        null=False,
        db_column='kind'
    )

    value = models.CharField(
        max_length=64,
        blank=False,
        null=False,
        db_column='value'
    )

    def save(self, *args, **kwargs):
        if self._state.adding:
            self.realm_id = self.subject.realm_id
        return super().save(*args, **kwargs)

    class Meta:
        app_label = 'iam'
        db_table = 'principals'
        default_permissions = []

        #: TODO: Worst normalization ever. A both realms and subjects can not
        #: have duplicate principals, but no composite keys in Django FTW.
        unique_together = [
            ('realm', 'kind', 'value'),
            ('subject', 'kind', 'value')
        ]




"""Declares :class:`SubjectRepository`."""
from django.apps import apps


class SubjectRepository:
    """Provides an interface to lookup :class:`~iam.ext.django.models.Subject`
    Data Access Objects (DAOs).
    """

    @property
    def model(self):
        return apps.get_model('iam.Subject')

    def get(self, pk):
        """Look up a :term:`Subject` by its identifier."""
        return self.model.objects.get(pk=pk)



#
# Copyright (C) 2019-2020 Cochise Ruhulessin
#
# This file is part of python-iam.
#
# python-iam is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# python-iam is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with python-iam.  If not, see <https://www.gnu.org/licenses/>.


TEST_USER1_ID           = '00000000-0000-0000-0000-000000000001'
TEST_USER1_EMAIL        = 'testuser1@test.iam.unimatrixone.io'
TEST_USER1_PHONENUMBER  = "+31612345678"
TEST_PASSWORD           = 'password'


MESSAGE_QUERY_PARAM = 'm'

MESSAGE_TOKEN_SESSION_KEY = 'iam.token'



#
# Copyright (C) 2019-2020 Cochise Ruhulessin
#
# This file is part of python-iam.
#
# python-iam is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# python-iam is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with python-iam.  If not, see <https://www.gnu.org/licenses/>.
import uuid

from .principal import Principal


class PrincipalFactory:
    """Creates :class:`~iam.domain.principal.Principal`
    instances from Django objects.
    """
    model = Principal

    def fromdjangomodel(self, resource):
        pk = resource.pk
        meta = resource._meta
        if not hasattr(resource, 'kind'):
            raise TypeError("Principal must expose a `kind` attribute.")
        if not isinstance(pk, (int, uuid.UUID))\
        and not hasattr(resource, 'get_resource_id'):
            raise ValueError(
                "Cannot determine resource identifier from primary key.")
        if hasattr(resource, 'get_resource_id'):
            resource_id = resource.get_resource_id()
        elif isinstance(pk, int):
            resource_id = uuid.UUID(int=pk)
        elif isinstance(pk, uuid.UUID):
            resource_id = pk
        return self.model(resource.kind, resource_id)



"""Declares the :class:`Subject` database object."""
import uuid

import ioc
from django.conf import settings
from django.utils.crypto import salted_hmac
from django.contrib.auth.base_user import AbstractBaseUser
from django.contrib.auth.hashers import is_password_usable
from django.core.exceptions import ObjectDoesNotExist
from django.db import models

from iam.lib.meta import AttributeDispatcher
from .platformagreements import PlatformAgreement
from .realms import Realm
from .passwordcredentials import PasswordCredential
from .permissionsmixin import PermissionsMixin


class Subject(AbstractBaseUser, PermissionsMixin):
    """A :class:`Subject` represents the source of a request to a system."""
    REQUIRED_FIELDS = ['realm']
    USERNAME_FIELD = 'id'
    credentials = ioc.class_property('CredentialService')
    storage_kinds = {
        'iam.unimatrixone.io/v1/FacebookAccountPrincipal': 'id.iam.unimatrixone.io/facebook',
        'iam.unimatrixone.io/v1/EmailPrincipal': 'id.iam.unimatrixone.io/email',
        'iam.unimatrixone.io/v1/GoogleAccountPrincipal': 'id.iam.unimatrixone.io/google',
        'iam.unimatrixone.io/v1/LinkedInAccountPrincipal': 'id.iam.unimatrixone.io/linkedin',
        'iam.unimatrixone.io/v1/NullPrincipal': 'id.iam.unimatrixone.io/null',
        'iam.unimatrixone.io/v1/PhonenumberPrincipal': 'id.iam.unimatrixone.io/phonenumber',
        'iam.unimatrixone.io/v1/TwitterAccountPrincipal': 'id.iam.unimatrixone.io/twitter',
    }

    # Remove properties that we don't want.
    password = last_login = None

    #: Maps :term:`Subject Properties` to related names.
    property_map = {
        'iam.unimatrixone.io/first-name'    : 'stringproperties',
        'iam.unimatrixone.io/last-name'     : 'stringproperties',
        'iam.unimatrixone.io/superuser'     : 'booleanproperties'
    }

    #: Indicates if the :class:`Subject` is authenticated. This attribute is
    #: always ``True``.
    is_authenticated = True

    #: Inverse of :attr:`is_authenticated`. This attribute is always ``False``.
    is_anonymous = False

    #: Specifies the :class:`~iam.ext.django.models.Realm` to which the
    #: :term:`Subject` belongs.
    realm = models.ForeignKey(
        Realm,
        on_delete=models.PROTECT,
        default=settings.IAM_REALM,
        blank=False,
        null=False,
        db_column='realm_name'
    )

    id = models.UUIDField(
        primary_key=True,
        blank=False,
        null=False,
        default=uuid.uuid4,
        db_column='id'
    )

    kind = models.CharField(
        max_length=63,
        blank=False,
        null=False,
        default='User',
        db_column='kind'
    )

    @property
    def first_name(self):
        return self._get_property('iam.unimatrixone.io/first-name')

    @first_name.setter
    def first_name(self, value):
        self._set_property('iam.unimatrixone.io/first-name', value)

    @property
    def is_active(self):
        return True

    @property
    def last_name(self):
        return self._get_property('iam.unimatrixone.io/last-name')

    @last_name.setter
    def last_name(self, value):
        self._set_property('iam.unimatrixone.io/last-name', value)

    @property
    def is_staff(self):
        return self.realm_id == settings.IAM_STAFF_REALM

    @property
    def is_superuser(self):
        return self._get_property('iam.unimatrixone.io/superuser')

    def accepts(self, agreement):
        """Return a boolean indicating if the :term:`Subject` accepted the
        :term:`Platform Agreement` `agreement`.
        """
        latest = PlatformAgreement.objects.effective(slug=agreement)
        qs = latest.languages.filter(accepted__subject__pk=self.pk)
        return qs.exists()

    def addcredential(self, credential):
        """Adds a :term:`Credential` to a :term:`Subject`.

        Args:
            credential (:class:`~iam.domain.Credential`): a credential
                implementation that the :term:`Subject` may use to prove its
                identity.

        Returns:
            :class:`types.NoneType`
        """
        assert type(credential).__name__ == 'PasswordCredential'
        dao, created = PasswordCredential.objects.get_or_create(
            subject=self, value=credential.secret
        )
        if not created:
            dao.value = credential.secret
            dao.save(update_fields=['value'])

    def addprincipal(self, principal):
        """Add a :term:`Principal` to the :term:`Subject`.

        Args:
            principal (:obj:`~iam.domain.Principal`): the :term:`Principal`
                to identify the :term:`Subject` with.

        Returns:
            :class:`types.NoneType`
        """
        return self.principals.create(kind=self.storage_kinds[principal.kind],
            value=principal.value).pk

    def check_credential(self, credential):
        """Verfifies the :term:`Credential` using the configured credential
        verification service.
        """
        if not credential:
            return False

        return self.credentials.verify(credential, self)

    def haspassword(self):
        """Return a boolean indicating if the :term:`Subject` has a
        :term:`Valid Password`.
        """
        try:
            self.password
            return True
        except (AttributeError, ObjectDoesNotExist):
            return False

    def hasprincipal(self, principal):
        """Return a boolean indicating if the :term:`Subject` is identified
        by the given :term:`Principal`.

        Args:
            principal (:class:`~iam.domain.Principal`): the :term:`Principal`
                to lookup.

        Returns:
            :obj:`bool`
        """
        return self.principals.filter(
            kind=self.storage_kinds[principal.kind],
            value=principal.value).exists()

    def marksuperuser(self, enabled):
        """Marks the :term:`Subject` as a :term:`Super User`."""
        kind = 'iam.unimatrixone.io/superuser'
        self._set_property(kind, enabled)

    # The methods below override some Django behavior.
    def has_usable_password(self):
        """Return False if set_unusable_password() has been called for
        this user.
        """
        try:
            return is_password_usable(self.password.value)
        except ObjectDoesNotExist:
            return False

    def get_full_name(self):
        """Return a string holding the full name for this :term:`Subject`."""
        return f'{self.first_name} {self.last_name}'

    def get_session_auth_hash(self):
        """Return a HMAC of the password; used to invalidate sessions by
        Django.
        """
        # This method is overridden because Django expects the password attribute
        # to always have a value. In the IAM application access of this attribute
        # may cause RelatedObjectDoesNotExist.
        key_salt = "iam.ext.django.models.Subject.get_session_auth_hash"
        try:
            return salted_hmac(key_salt, self.password.value).hexdigest()
        except ObjectDoesNotExist:
            return salted_hmac(key_salt, '').hexdigest()

    def __str__(self):
        return str(self.pk)

    class SubjectManager(models.Manager):

        def resolve(self, realm, principal):
            """Return a :class:`Subject` instance using a
            :class:`~iam.domain.Principal`.
            """
            return self.get_by_principal(realm, principal)

        def get_by_principal(self, realm, principal):
            """Return a :class:`Subject` instance using a
            :class:`~iam.domain.Principal`.
            """
            return self.get(
                realm__name=realm,
                principals__value=principal.value,
                principals__kind=self.model.storage_kinds[principal.kind])

    objects = SubjectManager()

    class Meta:
        app_label = 'iam'
        db_table = 'subjects'
        default_permissions = []

    def _get_property(self, kind):
        try:
            related_name = self.property_map[kind]
            return getattr(self, related_name).get(kind=kind).value
        except ObjectDoesNotExist:
            return None

    def _set_property(self, kind, value):
        manager = getattr(self, self.property_map[kind])
        try:
            dao = manager.get(kind=kind)
            dao.value = value
            dao.save(update_fields=['value'])
        except ObjectDoesNotExist:
            manager.create(kind=kind, value=value)



"""Declares :class:`LoginCode`."""
import django.utils.crypto
from django.db import models

from unimatrix.lib import timezone

import iam.domain


class LoginCode(models.Model):
    """Maintains information about one-time login codes.

    Attributes:
        subject (:class:`~iam.ext.django.models.Subject`): the
            :term:`Subject` for which the code is used to authenticate.
        code (:obj:`str`): the temporary login code.
        attempts (:obj:`int`): number of attempts to authenticate on the
            :attr:`subject`.
        generated (:obj:`int`): the date/time at which the code was generated,
            in milliseconds since the UNIX epoch.
    """
    MAX_ATTEMPTS = 5

    #: Minimum lifetime of the login code, in milliseconds. A new  code is
    #: not generated before :attr:`MIN_LIFETIME` is expired (based on the
    #: value of :attr:`generated`).
    MIN_LIFETIME = 1000 * 60 * 10

    subject = models.UUIDField(
        primary_key=True,
        blank=False,
        null=False,
        db_column='subject'
    )

    code = models.CharField(
        max_length=127,
        default=iam.domain.LoginCode.new,
        blank=False,
        null=False,
        db_column='code'
    )

    attempts = models.PositiveIntegerField(
        default=0,
        blank=False,
        null=False,
        db_column='attempts'
    )

    generated = models.BigIntegerField(
        default=timezone.now,
        blank=False,
        null=False,
        db_column='generated'
    )

    def verify(self, code):
        """Verify that `code` is correct and increase :attr:`attempts`. If the
        maximum attempts treshold is reached, delete the object.
        """
        # Delete the LoginCode if the maximum attempts is reached and the
        # minimum lifetime is expired.
        if (self.attempts >= self.MAX_ATTEMPTS):
            if (timezone.now() - self.generated) > self.MIN_LIFETIME:
                self.delete()
            return False
        isvalid = django.utils.crypto.constant_time_compare(code, self.code)
        if not isvalid:
            self.attempts = models.F('attempts') + 1
            self.save(update_fields=['attempts'])
            self.refresh_from_db()
        else:
            self.delete()
        return isvalid

    class Meta:
        app_label = 'iam'
        db_table = 'logincodes'
        default_permissions = []



# Generated by Django 3.0.4 on 2020-03-19 03:56

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('iam', '0018_platformagreementlanguage'),
    ]

    operations = [
        migrations.AlterField(
            model_name='platformagreementlanguage',
            name='checksum',
            field=models.CharField(db_column='checksum', editable=False, max_length=128, unique=True),
        ),
    ]



"""Declares :class:`DebugView`."""
from django.views.generic import TemplateView
from django.utils.decorators import method_decorator

from .decorators import requireconsent
from .iamcontextmixin import IAMContextMixin



@method_decorator(requireconsent('terms'), 'dispatch')
@method_decorator(requireconsent('privacy'), 'dispatch')
class DebugView(IAMContextMixin, TemplateView):
    template_name = 'iam/debug.html.j2'
    splash_image = None



# Generated by Django 3.0.4 on 2020-03-07 00:21

from django.db import migrations


def create_default_realm(apps, schema_editor):
    Realm = apps.get_model('iam.Realm')
    Realm.objects.create(name='default', label='default')


def create_master_realm(apps, schema_editor):
    Realm = apps.get_model('iam.Realm')
    Realm.objects.create(name='master', label='Master')


class Migration(migrations.Migration):

    dependencies = [
        ('iam', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(create_default_realm),
        migrations.RunPython(create_master_realm),
    ]



#
# Copyright (C) 2019-2020 Cochise Ruhulessin
#
# This file is part of python-iam.
#
# python-iam is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# python-iam is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with python-iam.  If not, see <https://www.gnu.org/licenses/>.
"""The :mod:`iam.ext.django` package integrates the Python IAM framework
with Django.
"""


default_app_config = 'iam.ext.django.apps.IAMConfig'



# Generated by Django 3.0.4 on 2020-03-07 02:18

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('iam', '0003_emailprincipal'),
    ]

    operations = [
        migrations.AlterField(
            model_name='emailprincipal',
            name='subject',
            field=models.ForeignKey(db_column='subject_id', on_delete=django.db.models.deletion.CASCADE, related_name='email', to=settings.AUTH_USER_MODEL),
        ),
    ]



"""Declares :class:`RedirectingFormView`."""
from django import forms
from django.contrib.auth.decorators import login_required
from django.contrib.auth.views import REDIRECT_FIELD_NAME
from django.contrib.auth.views import SuccessURLAllowedHostsMixin
from django.utils.decorators import method_decorator
from django.utils.http import url_has_allowed_host_and_scheme
from django.views.generic import FormView
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_protect
from django.views.decorators.debug import sensitive_post_parameters


DECORATORS = [never_cache, csrf_protect, sensitive_post_parameters,
    login_required]


@method_decorator(DECORATORS, 'dispatch')
class RedirectingFormView(SuccessURLAllowedHostsMixin, FormView):
    """A :class:`django.views.generic.FormView` implementation that
    redirects back to the URL provided in the query parameters.
    """
    redirect_field_name = REDIRECT_FIELD_NAME

    class base_form(forms.Form):

        def __init__(self, *args ,**kwargs):
            self.request = kwargs.pop('request', None)
            super().__init__(*args, **kwargs)

    def get_form_kwargs(self):
        """Return the keyword arguments used to instantiate the form
        class.
        """
        kwargs = super().get_form_kwargs()
        kwargs.update({
            'request': self.request
        })
        return kwargs

    def get_success_url(self):
        """Return the user-originating redirect URL if it's safe. An
        unsafe URL indicates an attack probe.
        """
        redirect_to = self.request.POST.get(
            self.redirect_field_name,
            self.request.GET.get(self.redirect_field_name, '')
        )
        url_is_safe = url_has_allowed_host_and_scheme(
            url=redirect_to,
            allowed_hosts=self.get_success_url_allowed_hosts(),
            require_https=self.request.is_secure(),
        )
        return redirect_to if url_is_safe else ''



#
# Copyright (C) 2019-2020 Cochise Ruhulessin
#
# This file is part of python-iam.
#
# python-iam is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# python-iam is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with python-iam.  If not, see <https://www.gnu.org/licenses/>.
"""Declares :class:`~LinkedInAccountPrincipal`."""
from .externalidentifierprincipal import ExternalIdentifierPrincipal


class LinkedInAccountPrincipal(ExternalIdentifierPrincipal):
    """A :term:`Principal` that identifies a :term:`Subject` using a
    LinkedIn Account ID.
    """
    resource_class = 'LinkedInAccountPrincipal'
    platform = 'linkedin'

    @classmethod
    def fromoauth2(cls, response):
        """Instantiate a new :class:`LinkedInAccountPrincipal` from the OAuth2
        response received by LinkedIn.

        Args:
            response (:obj:`dict`): a dictionary holding the response data.

        Returns:
            :class:`LinkedInAccountPrincipal`
        """
        return cls(response['id'])

    def __init__(self, user_id):
        super().__init__()
        self._id = user_id

    @property
    def value(self):
        return self._id

    def __str__(self):
        return self._id







"""Declares :class:`PlatformAgreementLanguage`."""
import hashlib

import markdown
from django.db import models
from django.conf import settings
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django.utils.translation import get_language


class PlatformAgreementLanguage(models.Model):
    """A :term:`Platform Agreement` in a specific language."""

    agreement = models.ForeignKey(
        'iam.PlatformAgreement',
        on_delete=models.CASCADE,
        blank=False,
        null=False,
        related_name='languages',
        db_column='agreement_slug'
    )

    #: A SHA-256 hash of :attr:`content`.
    checksum = models.CharField(
        max_length=128,
        blank=False,
        null=False,
        editable=False,
        unique=True,
        db_column='checksum'
    )

    #: The language in which :attr:`content` is written.
    language = models.CharField(
        max_length=15,
        blank=False,
        null=False,
        choices=tuple(settings.LANGUAGES),
        db_column='language'
    )

    #: The encoding of :attr:`content`
    encoding = models.CharField(
        max_length=31,
        default='utf-8',
        blank=False,
        null=False,
        editable=False,
        db_column='encoding'
    )

    #: The markup language of :attr:`content`.
    markup = models.CharField(
        max_length=31,
        default='markdown',
        blank=False,
        null=False,
        editable=False,
        db_column='markup'
    )

    #: The title of the translated agreemend.
    title = models.CharField(
        max_length=127,
        blank=False,
        null=False,
        db_column='title'
    )

    #: The content of the translated agreement.
    content = models.TextField(
        blank=False,
        null=False,
        db_column='content'
    )

    def generate_checksum(self):
        """Make a checksum of the agreements' content."""
        h = hashlib.sha256()
        h.update(str.encode(self.agreement_id))
        h.update(str.encode(self.language))
        h.update(str.encode(self.title, self.encoding))
        h.update(str.encode(self.content, self.encoding))
        return h.hexdigest()

    def has_agreements(self):
        """Return a boolean indicating if the :term:`Platform Agreement`
        has agreements from :term:`Subjects` in :attr:`language`.
        """
        return False

    def render(self):
        """Renders :attr:`content` in the specified :attr:`markup`."""
        assert self.markup == 'markdown'
        return markdown.markdown(self.content)

    def save(self, *args, **kwargs):
        """Ensure that the :attr:`checksum` field is updated and proceed to
        persist the entity.
        """
        if self.has_agreements():
            raise ValueError(
                "Can not modify agreements to which Subjects consented.")
        self.checksum = self.generate_checksum()
        update_fields = kwargs.get('update_fields')
        if update_fields:
            # Assume this is a list.
            update_fields.append('checksum')
        return super().save(*args, **kwargs)

    class PlatformAgreementLanguageManager(models.Manager):

        def get_effective(self, agreement):
            """Returns the :class:`PlatformAgreementLanguage` instance
            for the current language and specified :term:`Platform Agreement`
            `agreement`.
            """
            try:
                return self.get(
                    agreement_id=agreement,
                    agreement__effective__lte=timezone.now(),
                    language=get_language())
            except self.model.DoesNotExist:
                return self.get(
                    agreement_id=agreement,
                    agreement__effective__lte=timezone.now(),
                    language=settings.LANGUAGE_CODE)

    objects = PlatformAgreementLanguageManager()

    class Meta:
        app_label = 'iam'
        db_table = 'platformagreementlanguages'
        default_permissions = []
        verbose_name = _("Platform Agreement Translation")
        verbose_name_plural = _("Platform Agreements Translations")
        unique_together = ['agreement', 'language']



"""Declares :class:`RegistrationService`."""
import ioc
from django.apps import apps
from django.db import transaction


class RegistrationService:
    """Provides an interface to register new :term:`Subjects`."""
    availability = ioc.class_property('PrincipalAvailabilityService')

    def register(self, host, platform, realm, principal, credentials=None):
        """Register a new :term:`Subject`, that is identified by the given
        :term:`Principal` `principal` and authenticates with the
        :term:`Credentials` `credentials`. Return an :class:`uuid.UUID` that
        identifies the :term:`Subject`.
        """
        if not self.availability.isavailable(realm, principal):
            raise ValueError(
                f"{principal.kind}:{principal.value} is already associated.")
        Subject = apps.get_model('iam.Subject')
        SubjectRegistration = apps.get_model('iam.SubjectRegistration')
        with transaction.atomic():
            obj = Subject.objects.create(realm_id=realm)
            for credential in (credentials or []):
                obj.addcredential(credential)
            SubjectRegistration.objects.create(
                using_id=obj.addprincipal(principal),
                host=host,
                platform=platform
            )
        return obj.pk



"""Specifies the URLs for the :mod:`iam` demo project."""
from django.contrib import admin
from django.urls import include
from django.urls import path
from django.conf.urls.i18n import i18n_patterns
from django.conf.urls.static import static
from django.conf import settings

from ..django import views


urlpatterns = [
    path('auth/', include('social_django.urls',
        namespace=settings.SOCIAL_AUTH_URL_NAMESPACE)),
]

if settings.DEBUG:
    urlpatterns += static(settings.STATIC_URL, document_root=settings.STATIC_ROOT)

urlpatterns += i18n_patterns(
    path('admin/login/', views.LoginView.as_view(
        allow_signup=False,
        allow_local=False,
        realm=settings.IAM_STAFF_REALM,
        allow_public=False
    )),
    path('admin/', admin.site.urls),
    path('', include('iam.ext.django.urls',
        namespace=settings.IAM_URL_NAMESPACE))
)

admin.autodiscover()



# Generated by Django 3.0.4 on 2020-03-19 05:10

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('iam', '0020_acceptedplatformagreement'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='platformagreement',
            options={'default_permissions': [], 'ordering': ['-effective'], 'verbose_name': 'Platform Agreement', 'verbose_name_plural': 'Platform Agreements'},
        ),
        migrations.AlterField(
            model_name='acceptedplatformagreement',
            name='agreement',
            field=models.ForeignKey(db_column='agreement_checksum', on_delete=django.db.models.deletion.PROTECT, related_name='accepted', to='iam.PlatformAgreementLanguage', to_field='checksum'),
        ),
        migrations.AlterField(
            model_name='acceptedplatformagreement',
            name='subject',
            field=models.ForeignKey(db_column='subject_id', on_delete=django.db.models.deletion.CASCADE, related_name='agreements', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='platformagreementlanguage',
            name='agreement',
            field=models.ForeignKey(db_column='agreement_slug', on_delete=django.db.models.deletion.CASCADE, related_name='languages', to='iam.PlatformAgreement'),
        ),
    ]



"""Declares :class:`AcceptAgreementFormView`."""
import ioc
from django import forms
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse
from django.utils.decorators import method_decorator

from .basepartialform import BasePartialFormView
from ..models import PlatformAgreementLanguage


@method_decorator(login_required, 'dispatch')
class AcceptAgreementFormView(BasePartialFormView):
    template_name = 'iam/includes/forms/login.completenames.html.j2'
    agreement = ioc.class_property('AgreementAcceptanceService')

    class form_class(forms.Form):
        agreement = forms.ModelChoiceField(
            required=True,
            queryset=PlatformAgreementLanguage.objects.all(),
            to_field_name='checksum'
        )

    def form_invalid(self, form):
        """Indicate the caller that the form is not valid."""
        return HttpResponse(status=422)

    def form_valid(self, form):
        """Mark the agreement as accepted."""
        self.agreement.accept(form.cleaned_data['agreement'].checksum,
            self.request.user.pk, self.request.host)
        return HttpResponse(status=201)



"""Declares :class:`IAMContextMixin`."""
import os

import ioc

from django.conf import settings
from django.utils.translation import gettext_lazy as _
from django.contrib.auth.views import LoginView

from iam.const import MESSAGE_TOKEN_SESSION_KEY
from iam.ext.django.models import Realm


class IAMContextMixin:
    allow_signup = False
    allow_local = True
    redirect_field_name = LoginView.redirect_field_name
    principals = {
        'username': _('username'),
        'email': _('email address'),
        'phonenumber': _('phonenumber')
    }
    jwt = ioc.class_property('JWTService')

    #: Specifies the realm to which the :term:`Subject` will login.
    realm = settings.IAM_REALM

    #: Specifies the splash image.
    splash_image = settings.IAM_IMAGE_SPLASH

    #: Indicates if public (i.e. from the internet) :term:`Subjects` may
    #: login.
    allow_public = False

    def get_enabled_principals(self):
        """Returns the principal kinds that are enabled."""
        return list(settings.IAM_ACCOUNT_PRINCIPALS)

    def get_state_param(self):
        self.request.session[MESSAGE_TOKEN_SESSION_KEY] = token =\
            bytes.hex(os.urandom(16))
        return bytes.decode(self.jwt.encrypt({
            'rlm': self.realm,
            'token': token
        }))

    def get_context_data(self, *args, **kwargs):
        """Return the template context dictionary to use when rendering
        templates.
        """
        ctx = super().get_context_data(*args, **kwargs)
        ctx['iam_state_param'] = self.get_state_param()
        ctx['iam_portal_name']  = "Unimatrix One"
        ctx['iam_url_namespace'] = settings.IAM_URL_NAMESPACE
        ctx['iam_urls'] = {
            'forms': {
                'names': f'{settings.IAM_URL_NAMESPACE}:forms:names',
                'terms': f'{settings.IAM_URL_NAMESPACE}:forms:terms',
            },
            'login': f'{settings.IAM_URL_NAMESPACE}:login',
            'login_token': f'{settings.IAM_URL_NAMESPACE}:login.token',
            'logout': f'{settings.IAM_URL_NAMESPACE}:logout',
            'register': f'{settings.IAM_URL_NAMESPACE}:register',
            'resetcredentials': f'{settings.IAM_URL_NAMESPACE}:resetcredentials',
        }
        ctx['social_auth_begin_url'] = f'{settings.SOCIAL_AUTH_URL_NAMESPACE}:begin'
        ctx.update({
            'iam_allow_signup': self.allow_signup,
            'iam_enable_local': self.allow_local\
                and bool(settings.IAM_ACCOUNT_PRINCIPALS),
            'iam_enable_magic_link': settings.IAM_ENABLE_MAGIC_LINK,
            'iam_enable_password_reset': settings.IAM_ENABLE_PASSWORD_RESET,
            'iam_image_logo': settings.IAM_IMAGE_LOGO,
            'iam_principals_placeholder': self.get_principal_placeholder(),
            'iam_public_realm': self.allow_public,
            'iam_redirect_field_name': self.redirect_field_name,
            'iam_require_username': settings.IAM_REQUIRE_USERNAME,
        })

        capabilities = ctx['iam_capabilities'] = []

        if self.splash_image is not None:
            ctx['iam_image_splash'] = self.splash_image

        # Determine login capabilities from settings and realm
        realm = Realm.objects.get(name=self.realm)
        for provider in realm.getssoproviders():
            capabilities.append(f'login:{provider}')

        # Add the next URL to the template context to external providers
        # can also login and redirect to the appropriate location.
        #raise Exception(self.request.GET['next'])

        return ctx

    def get_principal_placeholder(self):
        """Return the placeholder for the principal input field based on the
        configured principal types.
        """
        principals = [str(self.principals[x])
            for x in self.get_enabled_principals()]
        placeholder = ''
        if len(principals) == 1:
            placeholder = principals[0]
        else:
            placeholder += str.join(', ', principals[:-1])
            placeholder += ' ' + str(_('or')) + ' '
            placeholder += principals[-1]
        return str.title(placeholder[0]) + placeholder[1:]

    def get_realm(self):
        """Returns the :term:`Realm` for this controller."""
        return self.realm



"""Specifies the URLs exposed by the :mod:`iam.ext.django` application."""
from django.urls import include
from django.urls import path
from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.contrib.auth.views import LogoutView
from django.views.generic import TemplateView
from django.utils.translation import gettext_lazy as _

from . import views


app_name = 'iam'

urlpatterns = []
if settings.DEBUG or settings.IAM_ENABLE_DEBUG_PAGE:
    handler = views.DebugView.as_view()
    urlpatterns += [
        path(str.lstrip(settings.IAM_DEBUG_PATH, '/'),
            login_required(handler), name='debug')
    ]


urlpatterns.extend([
    path('forms/',
        include(([
            path('names', views.ChooseNamesFormView.as_view(), name='names'),
            path('accept/terms', views.AcceptAgreementFormView.as_view(),
                name='terms'),
        ], 'iam'), namespace='forms'),
    ),
    path(_('login/'), views.LoginView.as_view(allow_public=True), name='login'),
    path(_('login/magic'), views.MagicLinkView.as_view(), name='login.magic'),
    path(_('login/token'), views.LoginTokenView.as_view(), name='login.token'),
    path(_('logout'), LogoutView.as_view(), name='logout'),
    path(_('register'), views.RegisterView.as_view(), name='register'),
])

if settings.IAM_ENABLE_PASSWORD_RESET:
    urlpatterns.append(
        path(_('login/reset'), views.ResetCredentialsView.as_view(),
            name='resetcredentials'))



#
# Copyright (C) 2019-2020 Cochise Ruhulessin
#
# This file is part of python-iam.
#
# python-iam is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# python-iam is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with python-iam.  If not, see <https://www.gnu.org/licenses/>.
from .abstractrealm import AbstractRealm
from .credential import Credential
from .emailprincipal import EmailPrincipal
from .facebookaccountprincipal import FacebookAccountPrincipal
from .googleaccountprincipal import GoogleAccountPrincipal
from .linkedinaccountprincipal import LinkedInAccountPrincipal
from .logincode import LoginCode
from .logincodecredential import LoginCodeCredential
from .nullcredential import NullCredential
from .nullprincipal import NullPrincipal
from .passwordcredential import PasswordCredential
from .phonenumberprincipal import PhonenumberPrincipal
from .principalfactory import PrincipalFactory
from .resource import Resource
from .resourcekind import ResourceKind
from .twitteraccountprincipal import TwitterAccountPrincipal



"""Declares :class:`OneTimePasswordService`."""
import ioc
from django.apps import apps
from unimatrix.lib import timezone


class OneTimePasswordService:
    """Provides an interface to generate and verify :term:`One Time Password`
    instances for a :term:`Subject`.
    """

    @property
    def model(self):
        return apps.get_model('iam.LoginCode')

    def generate(self, subject):
        """Generate a new :term:`One Time Password` for the :term:`Subject`
        identified by UUID `subject`. Return a boolean indicating if a new
        code was created.

        Args:
            subject (:class:`uuid.UUID`): identifies the :term:`Subject` to
                generate a code for.

        Returns:
            :obj:`bool`
        """
        # Ensure that expired login codes are deleted.
        self.model.objects.filter(subject=subject,
            generated__lte=(timezone.now() - self.model.MIN_LIFETIME)).delete()
        obj, created = self.model.objects.get_or_create(subject=subject)
        return obj.code

    def verify(self, subject, code):
        """Verify that `code` is valid for the given `subject`.

        Args:
            subject (:class:`uuid.UUID`): identifies the :term:`Subject` to
                verify a code for.
            code (:obj:`str`): the code to verify.

        Returns:
            :obj:`bool`
        """
        try:
            dao = self.model.objects.get(subject=subject)
        except self.model.DoesNotExist:
            return False
        return dao.verify(code)



import datetime
import re

import django.contrib.humanize.templatetags.humanize as humanize
import yaml
from django.contrib import messages
from django.contrib.staticfiles.storage import staticfiles_storage
from django.urls import reverse
from django.utils.translation import gettext, ngettext
from jinja2 import Environment


def naturaltime(value):
    if isinstance(value, int):
        if value == 0:
            value = 1
        value = datetime.datetime.fromtimestamp(value / 1000)
    return humanize.naturaltime(value)


def isactive(request, pattern):
    """Return a boolean indicating if a view with the given name
    is rendering the page.
    """
    return re.match(pattern, request.resolver_match.view_name)


def environment(**options):
    options.update({'extensions':['jinja2.ext.i18n']})
    env = Environment(**options)
    env.install_gettext_callables(gettext=gettext,
        ngettext=ngettext, newstyle=True)
    env.globals.update({
        'isactive': isactive,
        'static': staticfiles_storage.url,
        'url': reverse,
        'get_messages': messages.get_messages
    })
    env.filters.update({
        'naturaltime': naturaltime,
        'to_yaml': (lambda x: yaml.safe_dump(x, indent=2, default_flow_style=False)),
    })
    return env




"""Bootstraps the :mod:`iam.ext.django` Django application."""
import ioc
from django.apps import AppConfig

from . import infra
from . import services


class IAMConfig(AppConfig):
    name = 'iam.ext.django'
    label= 'iam'

    def ready(self):
        ioc.provide('AgreementAcceptanceService',
            services.AgreementAcceptanceService())
        ioc.provide('CredentialService', services.CredentialService())
        ioc.provide('HashedPasswordRepository',
            infra.HashedPasswordRepository())
        ioc.provide('JWTService', services.JWTService())
        ioc.provide('OneTimePasswordService',
            services.OneTimePasswordService())
        ioc.provide('PasswordService',
            services.PasswordService())
        ioc.provide('PasswordResetService',
            services.PasswordResetService())
        ioc.provide('PrincipalAvailabilityService',
            services.PrincipalAvailabilityService())
        ioc.provide('RegistrationService', services.RegistrationService())
        ioc.provide('SubjectRepository', infra.SubjectRepository())
        ioc.provide('iam.TemplateService', services.TemplateService())





class ResourceKind(str):

    @classmethod
    def fromdjango(cls, obj):
        """Create a new :class:`ResourceKind` object
        from a Django model class or instance.
        """
        meta = obj._meta
        return cls(f'{meta.app_label}.{meta.object_name}')



"""Declares :class:`PolymorphicPrincipalForm`."""
import ioc
from django import forms
from django.conf import settings
from django.contrib.auth import authenticate
from django.core.validators import MinLengthValidator
from django.utils.translation import gettext_lazy as _

from iam.domain import EmailPrincipal
from iam.domain import PhonenumberPrincipal
from iam.domain import NullPrincipal
from iam.domain import PasswordCredential
from iam.ext.django.models import Subject


class BasePrincipalForm(forms.Form):
    """Base class for all forms that need to guess principals."""

    #: A list of principal classes that are tried by the login form based
    #: in the `principal` input field. The first match is used.
    principal_classes = {
        'email': EmailPrincipal,
        'phonenumber': PhonenumberPrincipal,
    }

    principal = forms.CharField(
        max_length=128,
        required=True
    )

    def __init__(self, *args, **kwargs):
        self.realm = kwargs.pop('realm')
        self.request = kwargs.pop('request')
        self.subject = None
        super().__init__(*args, **kwargs)

    def clean_principal(self):
        """Resolves the provided principal to a concrete type."""
        if self.cleaned_data.get('principal'):
            self.guessed = self.guess_principal(self.cleaned_data['principal'])
            if type(self.guessed) == NullPrincipal:
                raise forms.ValidationError(f"Invalid input.", 'principal')
        return self.cleaned_data.get('principal')

    def get_principal_classes(self):
        """Returns the enabled principal classes."""
        return [self.principal_classes[x]
            for x in settings.IAM_ACCOUNT_PRINCIPALS] + [NullPrincipal]

    def get_principal(self, resolve=False):
        """Return a :class:`~iam.domain.Principal` instance using the value
        provided in the ``principal`` input field.
        """
        subject = None
        principal = self.guess_principal(self.cleaned_data['principal'])
        if resolve:
            try:
                self.subject = subject = principal.resolve(self.realm,
                    Subject.objects)
            except Subject.DoesNotExist:
                self.subject = subject = None
        return principal, subject

    def get_subject(self):
        """Return a :class:`~iam.ext.django.models.Subject` instance,
        identified by the :term:`Principal` from the input field.
        """
        if self.subject is not None:
            return self.subject
        _, self.subject = self.get_principal(resolve=True)
        return self.subject

    def get_subject_id(self):
        """Return the :term:`Subject Identifier`."""
        return self.get_subject().pk

    def get_user(self):
        """Returns the user found by the valid login credentials."""
        return self.subject

    def guess_principal(self, value):
        """Pattern matches a :term:`Principal Type` based on `value`."""
        for cls in self.get_principal_classes():
            if not cls.iscapable(value):
                continue
            break
        return cls(value)

class PolymorphicPrincipalForm(BasePrincipalForm):
    """Mixin class that determines the principal based on pattern
    matching, and retrieves the subject.
    """
    availability = ioc.class_property('PrincipalAvailabilityService')
    registration = ioc.class_property('RegistrationService')
    repo = ioc.class_property('SubjectRepository')

    #: Indicate that the form may create a new
    #: :class:`~iam.ext.django.models.Subject` based on the form parameters,
    #: if there was none prior existing.
    allow_create = False

    #: Indicate that the provided principal, password must authenticate
    #: to a :term:`Subject`. If :attr:`require_valid_credentials` is ``False``,
    #: then :meth:`get_user()` will return ``None`` instead of raising a
    #: :exc:`~django.forms.ValidationError` during form validation.
    require_valid_credentials = True

    password = forms.CharField(
        max_length=64,
        required=True,
        widget=forms.PasswordInput,
        validators=[
            MinLengthValidator(8)
        ]
    )

    def clean(self):
        """Looks up the subject specified by the input credential and
        raises a :exc:`ValidationError` if no subject could be
        authenticated. On success, add the
        :class:`~iam.ext.django.models.Subject` instance to the cleaned
        data.
        """
        assert not self.request.user.is_authenticated,\
            "Subject should not be authenticated when using this form."
        cleaned_data = super().clean()
        if not self.is_valid():
            return cleaned_data

        cleaned_data['realm'] = self.realm
        if cleaned_data.get('principal') and cleaned_data.get('password'):
            principal = self.guess_principal(cleaned_data['principal'])
            password = PasswordCredential(cleaned_data['password'])
            self.subject = authenticate(self.request, realm=self.realm,
                principal=principal, credential=password)
            if not self.subject:
                available = self.availability.isavailable(self.realm, principal)
                if self.require_valid_credentials and not available:
                    raise forms.ValidationError(
                        _("The login credentials that you provided are not "
                          "valid."))
                elif self.allow_create and available:
                    self.subject = self.create_subject(principal, password)
                elif not available:
                    # A Subject can be identified with the provided Principal,
                    # but its Credential was not valid.
                    self.fake_success = True
                else:
                    # The Principal is available, but we are not allowed
                    # to create Subjects.
                    raise forms.ValidationError(
                        _("The login credentials that you provided are not "
                          "valid."))

                # If no Subject was found at this point, we are an anonymous
                # user.
                if self.subject is None:
                    raise forms.ValidationError(
                        _("The login credentials that you provided are not "
                          "valid."))

            cleaned_data['subject'] = self.subject
            self.using = principal.kind

        return cleaned_data

    def create_subject(self, principal, password):
        """Creates a new :term:`Subject` that is identified by the given
        :term:`Principal` `principal` and :term:`Password` `password`. Return
        the :class:`~iam.ext.django.models.Subject` instance, or ``None``
        when the :term:`Principal` is already associated to a :term:`Subject`.
        """
        assert self.allow_create
        self.registration.register(self.request.host, self.request.site.domain,
            self.realm, principal, credentials=[password])
        return authenticate(self.request, realm=self.realm, principal=principal,
            credential=password)


class ChoosePasswordForm(forms.Form):
    password = ioc.class_property('PasswordService')

    password1 = forms.CharField(
        max_length=64,
        required=True,
        widget=forms.PasswordInput,
        validators=[
            MinLengthValidator(8)
        ]
    )
    password2 = forms.CharField(
        max_length=64,
        required=True,
        widget=forms.PasswordInput
    )

    def __init__(self, *args, **kwargs):
        self.realm = kwargs.pop('realm')
        self.request = kwargs.pop('request')
        super().__init__(*args, **kwargs)

    def clean(self):
        """Invoke the superclass :meth:`clean()` method and ensure that
        the passwords are equal.
        """
        data = super().clean()
        if not self.is_valid():
            return data

        if data['password1'] != data['password2']:
            self.add_error('password2', _("The passwords do not match."))
            return data

        # TODO: This needs to be invoked in the controller
        self.password.change(self.request.user.pk, data['password1'])

        return data

    def get_password(self):
        """Return the password provided by the user."""
        return self.cleaned_data['password1']

    def get_user(self):
        """Return the :term:`Subject` for which the :term:`Password` is being
        chosen.
        """
        return self.request.user



"""Declares :class:`AcceptedPlatformAgreement`."""
import hashlib

from django.db import models
from django.utils.translation import gettext_lazy as _
from unimatrix.lib import timezone


class AcceptedPlatformAgreement(models.Model):
    """Maintains information about :term:`Platform Agreements` that are
    accepted by a :term:`Subject`.
    """

    #: The :class:`~iam.ext.django.models.PlatformAgreementLanguage` that
    #: was accepted by :attr:`subject`.
    agreement = models.ForeignKey(
        'iam.PlatformAgreementLanguage',
        to_field='checksum',
        on_delete=models.PROTECT,
        blank=False,
        null=False,
        related_name='accepted',
        db_column='agreement_checksum'
    )

    #: The :class:`~iam.ext.django.models.Subject` that accepted
    #: :attr:`agreement`.
    subject = models.ForeignKey(
        'iam.Subject',
        on_delete=models.CASCADE,
        blank=False,
        null=False,
        related_name='agreements',
        db_column='subject_id'
    )

    #: The host address from which the :attr:`agreement` was accepted.
    host = models.GenericIPAddressField(
        blank=False,
        null=False,
        db_column='host'
    )

    #: The date/time at which the :attr:`agreement` was accepted, in
    #: milliseconds since the UNIX epoch.
    timestamp = models.BigIntegerField(
        default=timezone.now,
        blank=False,
        null=False,
        db_column='timestamp'
    )

    #: A SHA-256 hash of :attr:`content`.
    checksum = models.CharField(
        max_length=128,
        blank=False,
        null=False,
        editable=False,
        unique=True,
        db_column='checksum'
    )

    def generate_checksum(self):
        """Make a checksum of the agreements' content."""
        h = hashlib.sha256()
        h.update(str.encode(self.agreement_id))
        h.update(self.subject_id.bytes)
        h.update(int.to_bytes(self.timestamp, 8, 'big'))
        h.update(str.encode(self.host))
        return h.hexdigest()

    def save(self, *args, **kwargs):
        """Ensure that the :attr:`checksum` field is updated and proceed to
        persist the entity.
        """
        if not self._state.adding:
            raise ValueError("AcceptedPlatformAgreement objects are immutable.")
        self.checksum = self.generate_checksum()
        update_fields = kwargs.get('update_fields')
        if update_fields:
            update_fields.append('checksum')
        return super().save(*args, **kwargs)

    class Meta:
        app_label = 'iam'
        db_table = 'acceptedplatformagreements'
        default_permissions = []
        unique_together = ['agreement', 'subject']
        verbose_name = _('Accepted Platform Agreement')
        verbose_name_plural = _('Accepted Platform Agreements')



import base64
import json

import ioc
from django.conf import settings
from django.contrib.auth import logout
from django.contrib import messages
from django.db import transaction
from django.http import HttpResponseRedirect
from django.utils.translation import gettext as _
from social_core.exceptions import AuthAlreadyAssociated
from social_core.exceptions import AuthException
from social_core.exceptions import AuthForbidden

import iam.lib.urls
from iam.canon.exc import LoginAborted
from iam.const import MESSAGE_QUERY_PARAM
from iam.domain import FacebookAccountPrincipal
from iam.domain import EmailPrincipal
from iam.domain import GoogleAccountPrincipal
from iam.domain import LinkedInAccountPrincipal
from iam.domain import TwitterAccountPrincipal
from iam.lib.meta import AttributeDispatcher
from iam.ext.django.models import Realm
from iam.ext.django.models import Subject
from iam.ext.django.models import SubjectRegistration


def on_redirected(request, backend, *args, **kwargs):
    """Invoked directly when the SSO redirects back to our application."""
    response = None
    state = json.loads(bytes.decode(
        base64.urlsafe_b64decode(request.GET['state'])))
    if not state.get('payload'):
        request.realm = settings.IAM_REALM
    else:
        jwt = ioc.require('JWTService')
        payload, signed = jwt.decrypt(state['payload'])
        if not signed\
        or not payload\
        or (payload and signed and not jwt.verify(payload)):
            # This in all cases indicates tampering.
            payload = {'rlm': 'jail', 'reason': 'tamper'}
        else:
            payload = jwt.decode(payload)

        if payload.get('rlm') == 'jail':
            return HttpResponseRedirect(iam.lib.urls.reverse('login'))

        assert 'rlm' in payload
        request.realm = payload['rlm']

    realm = Realm.objects.get(name=request.realm)

    # Check if the realm allows login using this SSO.
    if not realm.allowsso(backend.name):
        messages.error(request, _("This provider is temporarily disabled."))
        return HttpResponseRedirect(iam.lib.urls.reverse('login'))


def on_authenticated(request, strategy, details, backend, *args, **kwargs):
    """Invoked when a user authenticates through the SSO provider."""
    resolver = IdentityProviderSubjectResolver()
    subject = None
    kwargs['request'] = request
    kwargs['realm'] = request.realm
    kwargs['site'] = request.site
    kwargs['host'] = request.host
    if request.user.is_authenticated:
        subject = request.user
    try:
        with transaction.atomic():
            uid, subject, new, social = resolver.resolve(
                backend, details, subject, *args, **kwargs)
    except LoginAborted as e:
        messages.error(request, e.reason)
        return HttpResponseRedirect(e.redirect_to)

    return {
        'social': social,
        'uid': uid,
        'user': subject,
        'is_new': new,
        'new_association': social is None
    }


class IdentityProviderSubjectResolver:

    @AttributeDispatcher.dispatch('name')
    def resolve(self, func, *args, **kwargs):
        return func(*args, **kwargs)

    @resolve.register('facebook')
    def resolve(self, backend, details, subject, response, is_new,
        *args, **kwargs):
        """Resolve a :term:`Subject` using a Google OAuth2 response.

        Args:
            backend (:class:`~social_core.backends.twitter.TwitterOAuth`): the
                :mod:`social_core` backend used to authenticate the subject.
            details (:obj:`dict`): a dictionary containing user details that
                were discovered by previous steps in the
                :term:`Authentication Pipeline`. In the default implementation,
                this is an empty dictionary.
            response (:obj:`dict`): the response that was received from Google.
            pipeline_index (:obj:`int`): indicates the position in the
                :term:`Authentication Pipeline` where this function was invoked.
                In the default implementation, `pipeline_index` is always ``0``.
            is_new (:obj:`bool`): indicates if the authenticated :term:`Subject`
                is a new user.
            **kwargs: keyword arguments not worthy of documentation.

        Returns:
            :obj:`tuple`
        """
        kwargs['platform'] = 'facebook'
        return self.process_response(FacebookAccountPrincipal.fromoauth2, backend,
            details, subject, response, is_new, *args, **kwargs)

    @resolve.register('google-oauth2')
    def resolve(self, backend, details, subject, response, is_new,
        *args, **kwargs):
        """Resolve a :term:`Subject` using a Google OAuth2 response.

        Args:
            backend (:class:`~social_core.backends.google.GoogleOAuth2`): the
                :mod:`social_core` backend used to authenticate the subject.
            details (:obj:`dict`): a dictionary containing user details that
                were discovered by previous steps in the
                :term:`Authentication Pipeline`. In the default implementation,
                this is an empty dictionary.
            response (:obj:`dict`): the response that was received from Google.
            pipeline_index (:obj:`int`): indicates the position in the
                :term:`Authentication Pipeline` where this function was invoked.
                In the default implementation, `pipeline_index` is always ``0``.
            is_new (:obj:`bool`): indicates if the authenticated :term:`Subject`
                is a new user.
            **kwargs: keyword arguments not worthy of documentation.

        Returns:
            :obj:`tuple`
        """
        if bool(response.get('email_verified')):
            kwargs['email'] = EmailPrincipal(response['email'])
        if response.get('hd'):
            kwargs['domain'] = response['hd']
        kwargs['first_name'] = response.get('given_name')
        kwargs['last_name'] = response.get('family_name')
        kwargs['platform'] = 'google'
        return self.process_response(GoogleAccountPrincipal.fromoauth2, backend,
            details, subject, response, is_new, *args, **kwargs)

    @resolve.register('linkedin-oauth2')
    def resolve(self, backend, details, subject, response, is_new,
        *args, **kwargs):
        """Resolve a :term:`Subject` using a LinkedIn OAuth2 response.

        Args:
            backend (:class:`~social_core.backends.twitter.TwitterOAuth`): the
                :mod:`social_core` backend used to authenticate the subject.
            details (:obj:`dict`): a dictionary containing user details that
                were discovered by previous steps in the
                :term:`Authentication Pipeline`. In the default implementation,
                this is an empty dictionary.
            response (:obj:`dict`): the response that was received from Google.
            pipeline_index (:obj:`int`): indicates the position in the
                :term:`Authentication Pipeline` where this function was invoked.
                In the default implementation, `pipeline_index` is always ``0``.
            is_new (:obj:`bool`): indicates if the authenticated :term:`Subject`
                is a new user.
            **kwargs: keyword arguments not worthy of documentation.

        Returns:
            :obj:`tuple`
        """
        kwargs['platform'] = 'linkedin'
        return self.process_response(LinkedInAccountPrincipal.fromoauth2, backend,
            details, subject, response, is_new, *args, **kwargs)

    @resolve.register('twitter')
    def resolve(self, backend, details, subject, response, is_new,
        *args, **kwargs):
        """Resolve a :term:`Subject` using a Google OAuth2 response.

        Args:
            backend (:class:`~social_core.backends.twitter.TwitterOAuth`): the
                :mod:`social_core` backend used to authenticate the subject.
            details (:obj:`dict`): a dictionary containing user details that
                were discovered by previous steps in the
                :term:`Authentication Pipeline`. In the default implementation,
                this is an empty dictionary.
            response (:obj:`dict`): the response that was received from Google.
            pipeline_index (:obj:`int`): indicates the position in the
                :term:`Authentication Pipeline` where this function was invoked.
                In the default implementation, `pipeline_index` is always ``0``.
            is_new (:obj:`bool`): indicates if the authenticated :term:`Subject`
                is a new user.
            **kwargs: keyword arguments not worthy of documentation.

        Returns:
            :obj:`tuple`
        """
        kwargs['platform'] = 'twitter'
        return self.process_response(TwitterAccountPrincipal.fromoauth2, backend,
            details, subject, response, is_new, *args, **kwargs)

    def process_response(self, principal_factory, backend, details, subject, response, is_new, *args, **kwargs):
        """Resolve a :term:`Subject` using an OAuth2 response.

        Args:
            backend (:class:`~social_core.backends.oauth.OAuthAuth`): the
                :mod:`social_core` backend used to authenticate the subject.
            details (:obj:`dict`): a dictionary containing user details that
                were discovered by previous steps in the
                :term:`Authentication Pipeline`. In the default implementation,
                this is an empty dictionary.
            response (:obj:`dict`): the response that was received from from
                the :term:`Identity Provider`.
            pipeline_index (:obj:`int`): indicates the position in the
                :term:`Authentication Pipeline` where this function was invoked.
                In the default implementation, `pipeline_index` is always ``0``.
            is_new (:obj:`bool`): indicates if the authenticated :term:`Subject`
                is a new user.
            **kwargs: keyword arguments not worthy of documentation.

        Returns:
            :obj:`tuple`
        """
        domain = kwargs.pop('domain', None)
        site = kwargs.pop('site')
        srcaddr = kwargs.pop('host')
        platform = kwargs.pop('platform')
        is_new = bool(kwargs.get('is_new'))
        realm = kwargs.pop('realm') or settings.IAM_REALM
        request = kwargs.pop('request')
        principal = principal_factory(response)
        social = backend.strategy.storage.user.get_social_auth(
            backend.name, principal.inrealm(realm))

        # TODO: All of this logic needs to be hidden behind service classes.
        config = settings.IAM_REALM_SETTINGS.get(realm)
        if not config.get('public'):
            # The realm is not public, check if the domain is whitelisted.
            if domain not in config.get('domains'):
                raise LoginAborted(
                    redirect_to=iam.lib.urls.reverse('login'),
                    reason=_("You are trying to access a protected area for "
                             "which you do not have access permissions."))

        # Check if a social account is associated to a user. If the user was
        # logged in and the social account is not yet associated, then
        # associate it with the user account.
        if social:
            if subject and social.user != subject:
                # Authenticated user that tries to associate a social account
                # that is already associated.
                raise AuthAlreadyAssociated("This account is already in use.")
            elif not subject:
                # Unauthenticated user that came in with a social account
                # that was previously associated.
                subject = social.user

        # Allow the lookup of subjects by some common principals. We have
        # validated and trust these principals.
        principals = [
            kwargs.pop('email', None)
        ]
        for p in principals:
            if p is None:
                continue
            try:
                subject = Subject.objects.get_by_principal(realm, p)
            except Subject.DoesNotExist:
                continue
            else:
                # Should not happen, but indicates that an other user has
                # claimed this principal.
                if social and (subject != social.user):
                    raise LoginAborted(
                        redirect_to=iam.lib.urls.reverse('login'),
                        reason=_("Unable to authenticate with the given credentials."))
                break

        # If subject is still None, create one.
        if subject is None:
            try:
                subject = principal.resolve(realm, Subject.objects)
            except Subject.DoesNotExist:
                is_new = True
                subject = Subject.objects.create(realm_id=realm)

                # TODO: Implement RegistrationService
                SubjectRegistration.objects.create(
                    using_id=subject.addprincipal(principal),
                    host=srcaddr,
                    platform=site.domain
                )

            assert subject.realm_id == realm

            # See if we can set up the account.
            if kwargs.get('first_name'):
                subject.first_name = kwargs.get('first_name')
            if kwargs.get('last_name'):
                subject.last_name = kwargs.get('last_name')

        elif subject and not social:
            subject.addprincipal(principal)

        if subject and subject.realm_id != realm:
            # Something went horribly wrong.
            raise NotImplementedError

        return principal.inrealm(realm), subject, is_new, social



# Generated by Django 3.0.4 on 2020-03-18 11:12

from django.db import migrations, models
import django.db.models.deletion
import unimatrix.lib.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('iam', '0011_auto_20200318_1042'),
    ]

    operations = [
        migrations.CreateModel(
            name='SubjectRegistration',
            fields=[
                ('using', models.OneToOneField(db_column='principal_id', on_delete=django.db.models.deletion.PROTECT, primary_key=True, serialize=False, to='iam.Principal')),
                ('host', models.GenericIPAddressField(db_column='host')),
                ('platform', models.CharField(db_column='platform', max_length=127)),
                ('registrar', models.CharField(db_column='registrar', default='self', max_length=127)),
                ('timestamp', models.BigIntegerField(db_column='timestamp', default=unimatrix.lib.timezone.now)),
                ('tsr', models.BinaryField(db_column='tsr', default=bytes)),
            ],
            options={
                'db_table': 'subjectregistrations',
                'default_permissions': [],
            },
        ),
    ]



"""Declares :class:`LoginTokenView`."""
import ioc
from django import forms
from django.contrib.auth import authenticate
from django.contrib.auth import login
from django.contrib.auth.views import LoginView
from django.http import HttpResponseRedirect
from django.utils.translation import gettext_lazy as _

import iam.lib.urls
from iam.domain import LoginCodeCredential
from .forms import BasePrincipalForm
from .subjectonboarding import SubjectOnboardingMixin


class LoginTokenView(SubjectOnboardingMixin, LoginView):
    """Provides handlers for ``GET`` and ``POST`` requests to login using
    a temporary token.
    """
    template_name = 'iam/login.token.html.j2'

    class form_class(BasePrincipalForm):
        token = forms.CharField(required=True, max_length=6)
        otp = ioc.class_property('OneTimePasswordService')

        def clean(self):
            """Invoke the superclass :meth:`clean()` method and use the valid
            form data to check if the login code is valid.
            """
            data = super().clean()
            if not self.is_valid():
                return data
            self.subject = authenticate(
                realm=self.realm,
                principal=self.guess_principal(data['principal']),
                credential=LoginCodeCredential(data['token']))
            if not self.subject:
                self.add_error('token',
                    _('The code you entered is not valid.'))
            return data

    def get_initial(self):
        """Parse initial data from the query parameters."""
        initial = super().get_initial()
        if self.request.method == 'GET':
            initial = {x: self.request.GET[x]
                for x in dict.keys(self.request.GET)}
        return initial

    def form_valid(self, form):
        """The login token was valid and the user is authenticated."""
        subject = form.get_user()
        request = self.request
        if subject.haspassword():
            return super().form_valid(form)

        login(self.request, subject)
        redirect_to = iam.lib.urls.reverse('login')
        if request.META['QUERY_STRING']:
            redirect_to += '?'
            redirect_to += request.META['QUERY_STRING']
        return HttpResponseRedirect(redirect_to)



"""Specifies configuration defaults for the :mod:`iam.ext.django` Django
application.
"""
import os

from django.urls import reverse_lazy
from unimatrix.lib.environ import parselist


__all__ = [
    'AUTH_USER_MODEL',
    'AUTHENTICATION_BACKENDS',
    'IAM_ACCOUNT_PRINCIPALS',
    'IAM_REALM',
    'IAM_REALM_SETTINGS',
    'IAM_DEBUG_PATH',
    'IAM_TERMS_OF_SERVICE',
    'IAM_PRIVACY_POLICY',
    'IAM_ENABLE_DEBUG_PAGE',
    'IAM_ENABLE_FACEBOOK',
    'IAM_ENABLE_GOOGLE',
    'IAM_ENABLE_LINKEDIN',
    'IAM_ENABLE_MAGIC_LINK',
    'IAM_ENABLE_PASSWORD_RESET',
    'IAM_ENABLE_TWITTER',
    'IAM_IMAGE_LOGO',
    'IAM_IMAGE_SPLASH',
    'IAM_REDIRECT_AUTHENTICATED_USER',
    'IAM_REQUIRE_USERNAME',
    'IAM_STAFF_REALM',
    'IAM_SUPERUSER_REALM',
    'IAM_URL_NAMESPACE',
    'LOGIN_URL',
    'LOGIN_REDIRECT_URL',
    'LOGOUT_REDIRECT_URL',
    'SOCIAL_AUTH_FACEBOOK_KEY',
    'SOCIAL_AUTH_FACEBOOK_SECRET',
    'SOCIAL_AUTH_GOOGLE_OAUTH2_KEY',
    'SOCIAL_AUTH_GOOGLE_OAUTH2_SECRET',
    'SOCIAL_AUTH_LINKEDIN_OAUTH2_KEY',
    'SOCIAL_AUTH_LINKEDIN_OAUTH2_SECRET',
    'SOCIAL_AUTH_LINKEDIN_OAUTH2_SCOPE',
    'SOCIAL_AUTH_LINKEDIN_OAUTH2_FIELD_SELECTORS',
    'SOCIAL_AUTH_LINKEDIN_OAUTH2_EXTRA_DATA',
    'SOCIAL_AUTH_PIPELINE',
    'SOCIAL_AUTH_TWITTER_KEY',
    'SOCIAL_AUTH_TWITTER_SECRET',
    'SOCIAL_AUTH_URL_NAMESPACE',
]

AUTH_USER_MODEL = 'iam.Subject'

AUTHENTICATION_BACKENDS = [
    'iam.ext.django.lib.backends.LocalSubjectBackend'
]


#: Account principal types that may be used to register a new account.
#: Valid values are 'username', 'email' and 'phonenumber'. May be provided
#: through the environment variable 'IAM_ACCOUNT_PRINCIPALS' as a
#: colon-separated list, using :mod:`iam.ext.django.defaults`. The ``email``
#: account principal is always enabled.
IAM_ACCOUNT_PRINCIPALS = parselist(os.environ, 'IAM_ACCOUNT_PRINCIPALS', cls=list)
if 'email' not in IAM_ACCOUNT_PRINCIPALS:
    IAM_ACCOUNT_PRINCIPALS.append('email')

IAM_DEBUG_PATH = os.getenv('IAM_DEBUG_PATH') or '/debug'

IAM_REALM = 'default'

#: Settings per realm.
IAM_REALM_SETTINGS = {}

#: Specifies the :term:`Realm` in which :term:`Super Users` are created.
IAM_SUPERUSER_REALM = os.getenv('IAM_SUPERUSER_REALM') or 'master'

#: Specifies the :term:`Realm` in which :term:`Staff Users` are created.
IAM_STAFF_REALM = os.getenv('IAM_STAFF_REALM') or 'backoffice'

IAM_ENABLE_DEBUG_PAGE = os.getenv('IAM_ENABLE_DEBUG_PAGE') == '1'

IAM_ENABLE_GOOGLE = os.getenv('IAM_ENABLE_GOOGLE') == '1'

IAM_ENABLE_FACEBOOK = os.getenv('IAM_ENABLE_FACEBOOK') == '1'

IAM_ENABLE_LINKEDIN = os.getenv('IAM_ENABLE_LINKEDIN') == '1'

IAM_ENABLE_MAGIC_LINK = os.getenv('IAM_ENABLE_MAGIC_LINK') == '1'

IAM_ENABLE_PASSWORD_RESET = os.getenv('IAM_ENABLE_PASSWORD_RESET') == '1'

IAM_ENABLE_TWITTER = os.getenv('IAM_ENABLE_TWITTER') == '1'

IAM_IMAGE_LOGO = os.getenv('IAM_IMAGE_LOGO')\
    or 'iam/logo.svg'

IAM_IMAGE_SPLASH = os.getenv('IAM_IMAGE_SPLASH')\
    or 'iam/splash.jpg'

#: Slug of the :class:`~iam.ext.django.models.PlatformAgremeent` that
#: maintains the :term:`Privacy Policy`.
IAM_PRIVACY_POLICY = 'privacy'


#: Redirect authenticated users when they visit the login page. Default is
#: ``True``.
IAM_REDIRECT_AUTHENTICATED_USER = os.getenv('IAM_REDIRECT_AUTHENTICATED_USER') == '1'
if os.getenv('IAM_REDIRECT_AUTHENTICATED_USER') is None:
    IAM_REDIRECT_AUTHENTICATED_USER = True


IAM_REQUIRE_USERNAME = os.getenv('IAM_REQUIRE_USERNAME') == '1'

#: Slug of the :class:`~iam.ext.django.models.PlatformAgremeent` that
#: maintains the :term:`Terms of Service`.
IAM_TERMS_OF_SERVICE = 'terms'

IAM_URL_NAMESPACE = 'iam'

LOGIN_URL = reverse_lazy(f'{IAM_URL_NAMESPACE}:login')

LOGIN_REDIRECT_URL = '/'

LOGOUT_REDIRECT_URL = reverse_lazy(f'{IAM_URL_NAMESPACE}:login')

SOCIAL_AUTH_FACEBOOK_KEY = os.getenv('IAM_FACEBOOK_KEY')

SOCIAL_AUTH_FACEBOOK_SECRET = os.getenv('IAM_FACEBOOK_SECRET')

SOCIAL_AUTH_GOOGLE_OAUTH2_KEY = os.getenv('GOOGLE_OAUTH2_KEY')

SOCIAL_AUTH_GOOGLE_OAUTH2_SECRET = os.getenv('GOOGLE_OAUTH2_SECRET')

SOCIAL_AUTH_LINKEDIN_OAUTH2_KEY = os.getenv('IAM_LINKEDIN_KEY')

SOCIAL_AUTH_LINKEDIN_OAUTH2_SECRET = os.getenv('IAM_LINKEDIN_SECRET')

SOCIAL_AUTH_LINKEDIN_OAUTH2_SCOPE = ['r_basicprofile', 'r_emailaddress']

SOCIAL_AUTH_LINKEDIN_OAUTH2_FIELD_SELECTORS = [
    'email-address',
    'formatted-name',
    'public-profile-url',
    'picture-url'
]

SOCIAL_AUTH_PIPELINE = (
    'iam.ext.django.lib.auth.on_redirected',
    'iam.ext.django.lib.auth.on_authenticated',
    'social_core.pipeline.social_auth.associate_user',
    'social_core.pipeline.social_auth.load_extra_data',
    'social_core.pipeline.user.user_details'
)

SOCIAL_AUTH_LINKEDIN_OAUTH2_EXTRA_DATA = [
    ('id', 'id'),
    ('formattedName', 'name'),
    ('emailAddress', 'email_address'),
    ('pictureUrl', 'picture_url'),
    ('publicProfileUrl', 'profile_url'),
]

SOCIAL_AUTH_TWITTER_KEY = os.getenv('IAM_TWITTER_KEY')

SOCIAL_AUTH_TWITTER_SECRET = os.getenv('IAM_TWITTER_SECRET')

SOCIAL_AUTH_URL_NAMESPACE = 'social'

if IAM_ENABLE_FACEBOOK:
    AUTHENTICATION_BACKENDS.insert(0,
        'social_core.backends.facebook.FacebookOAuth2')

if IAM_ENABLE_GOOGLE:
    AUTHENTICATION_BACKENDS.insert(0,
        'iam.lib.backends.google.GoogleOAuth2')

if IAM_ENABLE_LINKEDIN:
    AUTHENTICATION_BACKENDS.insert(0,
        'social_core.backends.linkedin.LinkedinOAuth2')

if IAM_ENABLE_TWITTER:
    AUTHENTICATION_BACKENDS.insert(0,
        'social_core.backends.twitter.TwitterOAuth')



"""Declares :class:`AcceptedPlatformAgreementAdmin`."""
from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from ..models import AcceptedPlatformAgreement
from .readonly import ReadonlyModelAdminMixin


class AcceptedPlatformAgreementAdmin(ReadonlyModelAdminMixin, admin.ModelAdmin):
    list_filter = ['subject__realm']
    list_display = ['timestamp', 'title', 'host', 'first_name', 'last_name']

    def title(self, obj):
        return f'{obj.agreement.title} {obj.agreement.agreement.effective.date()}'
    title.short_description = _('Title')

    def first_name(self, obj):
        return obj.subject.first_name or ''
    first_name.short_description = _('First name')

    def last_name(self, obj):
        return obj.subject.last_name or ''
    last_name.short_description = _('Last name')


admin.site.register(AcceptedPlatformAgreement, AcceptedPlatformAgreementAdmin)



# Generated by Django 3.0.4 on 2020-03-07 02:18

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('iam', '0002_auto_20200307_0021'),
    ]

    operations = [
        migrations.CreateModel(
            name='EmailPrincipal',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.EmailField(db_column='value', max_length=254)),
                ('realm', models.ForeignKey(db_column='realm_name', on_delete=django.db.models.deletion.PROTECT, to='iam.Realm')),
                ('subject', models.ForeignKey(db_column='subject_id', on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'emailprincipals',
                'default_permissions': [],
                'unique_together': {('realm', 'value'), ('subject', 'value')},
            },
        ),
    ]



"""Declares :class:`SubjectRegistration`."""
from django.db import models
from django.utils.translation import gettext as _

import unimatrix.lib.timezone


class SubjectRegistration(models.Model):
    """Maintains information about the registration of a :term:`Subject`."""

    #: The :term:`Principal` that the :term:`Subject` or its :term:`Registrar`
    #: used to create the account.
    using = models.OneToOneField(
        'iam.Principal',
        on_delete=models.PROTECT,
        blank=False,
        null=False,
        primary_key=True,
        db_column='principal_id'
    )

    #: The host address from which the :term:`Subject` was registered.
    host = models.GenericIPAddressField(
        blank=False,
        null=False,
        db_column='host'
    )

    #: Identifies the platform through which the :term:`Subject` registered.
    platform = models.CharField(
        max_length=127,
        blank=False,
        null=False,
        db_column='platform'
    )

    #: Identifies the entity that registered the :term:`Subject`. The value
    #: ``self`` indicates that the :term:`Subject` registered on its own
    #: behalf (e.g. through a publicly accessible web interface).
    registrar = models.CharField(
        max_length=127,
        default='self',
        blank=False,
        null=False,
        db_column='registrar'
    )

    #: The date/time at which the :term:`Subject` was registered, in
    #: milliseconds since the UNIX epoch.
    timestamp = models.BigIntegerField(
        default=unimatrix.lib.timezone.now,
        blank=False,
        null=False,
        db_column='timestamp'
    )

    #: Two zero octets followed by the DER-encoded response of a trusted
    #: :term:`Time Stamping Authority` (TSA) compliant to the :rfc:`3161`
    #: `Internet X.509 Public Key Infrastructure Time-Stamp Protocol (TSP)`
    #: specification.
    #:
    #: The ``hashedMessage`` is a SHA-256 hash consisting of the following
    #: data-elements:
    #:
    #: - :attr:`~iam.ext.django.models.Principal.subject_id` (from :attr:`using`)
    #: - :attr:`~iam.ext.django.models.Principal.kind` (from :attr:`using`)
    #: - :attr:`~iam.ext.django.models.Principal.value` (from :attr:`using`)
    #: - :attr:`host`
    #: - :attr:`using`
    #: - :attr:`registrar`
    #: - :attr:`timestamp`
    #:
    #: Since it is not assumed that a highly-available :term:`TSA` is deployed
    #: in every scenario, this field initializes to an empty byte-sequence and
    #: is processed asynchronously. Implementations should take care to ensure
    #: that the field is updated as soon as possible after registration. This
    #: may be done using the ``python manage.py updateregistrationtsrs``.
    tsr = models.BinaryField(
        default=bytes,
        blank=False,
        null=False,
        db_column='tsr'
    )

    class Meta:
        app_label = 'iam'
        db_table = 'subjectregistrations'
        default_permissions = []
        verbose_name = _("Registration")
        verbose_name_plural = _("Registrations")



# Generated by Django 3.0.4 on 2020-03-18 21:46

from django.db import migrations


def create_jail_realm(apps, schema_editor):
    Realm = apps.get_model('iam.Realm')
    Realm.objects.create(name='jail', label='Jail')


class Migration(migrations.Migration):

    dependencies = [
        ('iam', '0015_auto_20200318_1830'),
    ]

    operations = [
        migrations.RunPython(create_jail_realm)
    ]



"""Declares :class:`PlatformAgreement`."""
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _


class PlatformAgreement(models.Model):
    """Maintains a :term:`Agreement` for use of a system."""

    slug = models.SlugField(
        primary_key=True,
        blank=False,
        null=False,
        db_column='slug'
    )

    #: Human-readable title of the :term:`Platform Agreement`.
    title = models.CharField(
        max_length=127,
        blank=False,
        null=False,
        db_column='title'
    )

    #: The date/time at which the :term:`Platform Agreement` becomes
    #: effective.
    effective = models.DateTimeField(
        blank=False,
        null=False,
        db_column='effective'
    )

    def iseffective(self):
        """Return a boolean indicating if the :term:`Platform Agreement`
        is effective.
        """
        return self.effective <= timezone.now()

    class PlatformAgreementManager(models.Manager):

        def effective(self, slug):
            """Return the :term:`Effective Platform Agreement`, identified by
            the given `slug`.
            """
            return self.filter(slug=slug).order_by('-effective')[:1].get()

    objects = PlatformAgreementManager()

    class Meta:
        app_label = 'iam'
        db_table = 'platformagreements'
        default_permissions = []
        ordering = ['-effective']
        verbose_name = _("Platform Agreement")
        verbose_name_plural = _("Platform Agreements")



# Generated by Django 3.0.4 on 2020-03-19 04:29

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import unimatrix.lib.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('iam', '0019_auto_20200319_0356'),
    ]

    operations = [
        migrations.CreateModel(
            name='AcceptedPlatformAgreement',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('host', models.GenericIPAddressField(db_column='host')),
                ('timestamp', models.BigIntegerField(db_column='timestamp', default=unimatrix.lib.timezone.now)),
                ('checksum', models.CharField(db_column='checksum', editable=False, max_length=128, unique=True)),
                ('agreement', models.ForeignKey(db_column='agreement_checksum', on_delete=django.db.models.deletion.PROTECT, to='iam.PlatformAgreementLanguage', to_field='checksum')),
                ('subject', models.ForeignKey(db_column='subject_id', on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Accepted Platform Agreement',
                'verbose_name_plural': 'Accepted Platform Agreements',
                'db_table': 'acceptedplatformagreements',
                'default_permissions': [],
                'unique_together': {('agreement', 'subject')},
            },
        ),
    ]



"""Declares :class:`ResetCredentialsView`."""
import ioc
from django.contrib.auth import logout
from django.contrib.auth.views import LoginView
from django import forms

import iam.lib.urls
from .forms import BasePrincipalForm
from .subjectonboarding import SubjectOnboardingMixin


class ResetCredentialsView(SubjectOnboardingMixin, LoginView):
    """Provides ``GET`` and ``POST`` request handlers to reset the credentials
    for an account.
    """
    reset = ioc.class_property('PasswordResetService')
    template_name = 'iam/resetcredentials.html.j2'

    class form_class(BasePrincipalForm):

        principal = forms.CharField(
            max_length=64,
            required=True
        )

        def get_user(self):
            """Returns the user found by the valid login credentials."""
            return self.subject

    def form_valid(self, form):
        """We found to user and can now continue the flow."""
        logout(self.request)
        principal, subject = form.get_principal(resolve=True)
        if subject is not None:
            self.reset.challenge(principal,
                verify=self.get_token_login_url(),
                context_data={
                    'brand_name': self.request.site.name or self.request.site.domain
                })
        return self.render_to_response(
            self.get_context_data(form=form, state='challenged')
        )

    def get_token_login_url(self):
        """Return the URL at which the :term:`Subject` can use the
        reset token.
        """
        uri = self.request.build_absolute_uri(
            iam.lib.urls.reverse('login.token'))
        if self.request.META['QUERY_STRING']:
            uri += f"?{self.request.META['QUERY_STRING']}"
        return uri

    def get_template_names(self):
        """Return a list of template names."""
        if self.request.method == 'GET':
            return super().get_template_names()

        return ['iam/login.token.html.j2']



# Generated by Django 3.0.4 on 2020-03-07 03:42

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('iam', '0004_auto_20200307_0218'),
    ]

    operations = [
        migrations.CreateModel(
            name='PasswordCredential',
            fields=[
                ('subject', models.OneToOneField(db_column='subject_id', on_delete=django.db.models.deletion.CASCADE, primary_key=True, related_name='password', serialize=False, to=settings.AUTH_USER_MODEL)),
                ('value', models.CharField(db_column='value', max_length=128)),
            ],
            options={
                'db_table': 'passwordcredentials',
                'default_permissions': [],
            },
        ),
    ]



"""Declares :class:`SubjectRegistrationAdmin`."""
from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from iam.ext.django.models import SubjectRegistration
from .readonly import ReadonlyModelAdminMixin


class SubjectRegistrationAdmin(ReadonlyModelAdminMixin, admin.ModelAdmin):
    readonly = True
    search_fields = ['using__subject__stringproperties__value']
    list_filter = ['using__realm', 'registrar', 'platform']
    list_display = ['timestamp', 'realm', 'platform', 'registrar', 'using_short',
        'host', 'first_name', 'last_name']
    ordering = ['-timestamp']

    def using_short(self, obj):
        return str.split(obj.using.kind, '/')[-1]
    using_short.short_description = _('Using')

    def realm(self, obj):
        return obj.using.realm_id
    realm.short_description = _('Realm')

    def first_name(self, obj):
        return obj.using.subject.first_name or ''
    first_name.short_description = _('First name')

    def last_name(self, obj):
        return obj.using.subject.last_name or ''
    last_name.short_description = _('Last name')


admin.site.register(SubjectRegistration, SubjectRegistrationAdmin)



"""Declares :class:`ChooseNamesFormView`."""
from django import forms

from ..lib.forms import NameField
from .basepartialform import BasePartialFormView


class ChooseNamesFormView(BasePartialFormView):
    template_name = 'iam/includes/forms/login.completenames.html.j2'


    class form_class(forms.Form):
        first_name = NameField(
            required=True,
            max_length=32
        )

        last_name = NameField(
            required=True,
            max_length=32
        )

    def form_valid(self, form):
        """Update the :term:`Subject` with the given first and last name."""
        self.request.user.first_name = form.cleaned_data['first_name']
        self.request.user.last_name = form.cleaned_data['last_name']
        return super().form_valid(form)



#
# Copyright (C) 2019-2020 Cochise Ruhulessin
#
# This file is part of python-iam.
#
# python-iam is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# python-iam is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with python-iam.  If not, see <https://www.gnu.org/licenses/>.
"""Declares :class:`~EmailPrincipal`."""
import ioc

from .principal import Principal


class EmailPrincipal(Principal):
    """A :term:`Principal` that identifies a :term:`Subject` using an
    email address.
    """
    resource_class = 'EmailPrincipal'
    pattern = '(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)'

    @property
    def value(self):
        return self.email

    def __init__(self, email):
        super().__init__()
        self.email = email

    @ioc.inject('service', 'EmailService')
    def send(self, service, subject, text, html=None, sender=None):
        return service.send(sender, subject, [self.email], text, html)

    def __str__(self):
        return self.email or ''



from django.apps import apps


class HashedPasswordRepository:

    @property
    def model(self):
        return apps.get_model('iam.PasswordCredential')

    def get(self, subject):
        """Returns the hashed password that may be used to authenticate
        the given :term:`Subject` `subject`.

        Args:
            subject (:obj:`str`): the :term:`Subject Identifier` to look up
                the password for.

        Returns:
            :obj:`str` or ``None`` if the :term:`Subject` does not have a
            password.
        """
        try:
            return self.model.objects.get(subject_id=subject)
        except self.model.DoesNotExist:
            return None



# Generated by Django 3.0.4 on 2020-03-18 18:30

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('iam', '0014_subjectbooleanproperty'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='subjectregistration',
            options={'default_permissions': [], 'verbose_name': 'Registration', 'verbose_name_plural': 'Registrations'},
        ),
    ]



# Generated by Django 3.0.4 on 2020-03-18 23:40
import hashlib

from django.db import migrations, models
import django.db.models.deletion


LOREM_IPSUM = """

## Preamble

Could someone survive inside a transporter buffer for 75 years? Fear is the
true enemy, the only enemy. I suggest you drop it, Mr. Data. I think you've
let your personal feelings cloud your judgement. Flair is what marks the
difference between artistry and mere competence. Some days you get the bear,
and some days the bear gets you. Maybe if we felt any human loss as keenly as
we feel one of those close to us, human history would be far less bloody.

## An additional paragraph

We have a saboteur aboard. Earl Grey tea, watercress sandwiches... and
Bularian canapés? Are you up for promotion? Some days you get the bear,
and some days the bear gets you. My oath is between Captain Kargan and
myself. Your only concern is with how you obey my orders. Or do you prefer
the rank of prisoner to that of lieutenant? Why don't we just give everybody
a promotion and call it a night - 'Commander'? Fate. It protects fools,
little children, and ships named "Enterprise." We could cause a diplomatic
crisis. Take the ship into the Neutral Zone That might've been one of the
shortest assignments in the history of Starfleet.

## Another paragraph

I'll be sure to note that in my log. When has justice ever been as simple as
a rule book? Your shields were failing, sir. What's a knock-out like you doing
in a computer-generated gin joint like this? The Enterprise computer system
is controlled by three primary main processor cores, cross-linked with a
redundant melacortz ramistat, fourteen kiloquad interface modules. This is
not about revenge. This is about justice.

## Appendices

This section contains appendices.

### Appendix A

Did you come here for something in particular or just general Riker-bashing?
I'll alert the crew. I am your worst nightmare! Besides, you look good in a
dress. They were just sucked into space. Computer, lights up! Fate protects
fools, little children and ships named Enterprise.

### Appendix B

Maybe we better talk out here; the observation lounge has turned into a
swamp. Talk about going nowhere fast. and attack the Romulans. Well,
that's certainly good to know. Yesterday I did not know how to eat gagh.
Captain, why are we out here chasing comets?
"""


def create_lorem_terms(apps, schema_editor):
    PlatformAgreementLanguage = apps.get_model('iam.PlatformAgreementLanguage')
    PlatformAgreementLanguage.objects.create(
        agreement_id='terms',
        checksum=hashlib.sha256(b'1').hexdigest(),
        title="Terms of Service",
        language='en',
        content="# Terms of Service Placeholder\n" + LOREM_IPSUM
    )


def create_lorem_privacy(apps, schema_editor):
    PlatformAgreementLanguage = apps.get_model('iam.PlatformAgreementLanguage')
    PlatformAgreementLanguage.objects.create(
        agreement_id='privacy',
        checksum=hashlib.sha256(b'2').hexdigest(),
        title="Privacy Statement",
        language='en',
        content="# Privacy Statement Placeholder\n" + LOREM_IPSUM
    )


class Migration(migrations.Migration):

    dependencies = [
        ('iam', '0017_platformagreement'),
    ]

    operations = [
        migrations.CreateModel(
            name='PlatformAgreementLanguage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('checksum', models.CharField(db_column='checksum', unique=True, max_length=128)),
                ('language', models.CharField(choices=[('en', 'English'), ('nl', 'Dutch')], db_column='language', max_length=15)),
                ('encoding', models.CharField(db_column='encoding', default='utf-8', editable=False, max_length=31)),
                ('markup', models.CharField(db_column='markup', default='markdown', editable=False, max_length=31)),
                ('title', models.CharField(db_column='title', max_length=127)),
                ('content', models.TextField(db_column='content')),
                ('agreement', models.ForeignKey(db_column='agreement_slug', on_delete=django.db.models.deletion.CASCADE, to='iam.PlatformAgreement')),
            ],
            options={
                'verbose_name': 'Platform Agreement Translation',
                'verbose_name_plural': 'Platform Agreements Translations',
                'db_table': 'platformagreementlanguages',
                'default_permissions': [],
                'unique_together': {('agreement', 'language')},
            },
        ),
        migrations.RunPython(create_lorem_terms),
        migrations.RunPython(create_lorem_privacy),
    ]



"""Declares :class:`IAMMiddleware`."""
import ipaddress
import re

from django.contrib.auth import BACKEND_SESSION_KEY


class IAMMiddleware:
    """Sets various properties related to identity and access management
    on an incoming :term:`HTTP Request`.

    This request middleware requires session and must therefore appear after
    ``django.contrib.sessions.middleware.SessionMiddleware`` in the
    ``MIDDLEWARE`` setting.
    """

    #: Regular expression pattern that indicates if the backend used to
    #: authenticate was external (i.e. SSO).
    sso_backends_pattern = re.compile('^(social_core.backends|iam.lib.backends)\..*$')

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        # Set a flag indicating if the Subject authenticated through SSO.
        request.sso = False
        if request.session.get(BACKEND_SESSION_KEY):
            request.sso = bool(self.sso_backends_pattern.match(
                request.session[BACKEND_SESSION_KEY]))

        request.host = self.get_client_ip(request)

        return self.get_response(request)

    def get_client_ip(self, request):
        """Returns the remote address based on the ``HTTP_X_FORWARDED_FOR`` or
        ``REMOTE_ADDR`` headers.
        """
        remote_address = request.META.get('HTTP_CF_CONNECTING_IP')\
            or request.META.get('REMOTE_ADDR')
        x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
        if x_forwarded_for:
            proxies = [str.strip(x) for x in str.split(x_forwarded_for, ',')]
            remote_address = proxies[-1]

        assert remote_address is not None
        return remote_address



"""Declares :class:`SubjectStringProperty`."""
from django.db import models

from .subjectproperties import SubjectProperty


class SubjectStringProperty(SubjectProperty):
    """Maintains a :term:`Subject Property` of type :obj:`str`.

    Attributes:
        value (:obj:`str`): the value of the :term:`Subject Property`.
    """
    value = models.CharField(
        max_length=127,
        blank=False,
        null=False,
        db_column='value'
    )

    class Meta:
        app_label = 'iam'
        db_table = 'subjectstringproperties'
        default_permissions = []
        default_related_name = 'stringproperties'
        unique_together = ['subject', 'kind']



#pylint: skip-file
import unimatrix.lib.test
import django.test

from iam.const import TEST_USER1_ID
from ...models import LoginCode
from ..onetimepassword import OneTimePasswordService


@unimatrix.lib.test.integration
class OneTimePasswordAuthenticationTestCase(django.test.TestCase):
    fixtures = ['testusers']
    subject = TEST_USER1_ID

    def setUp(self):
        # Ensure that the minimum lifetime is not enforced.
        self.lifetime = LoginCode.MIN_LIFETIME
        LoginCode.MIN_LIFETIME = 0
        self.service = OneTimePasswordService()
        self.service.generate(self.subject)

    def tearDown(self):
        LoginCode.MIN_LIFETIME = self.lifetime

    def test_valid_code_can_authenticate(self):
        dao = LoginCode.objects.get(subject=self.subject)
        self.assertTrue(self.service.verify(self.subject, dao.code))

    def test_invalid_code_can_not_authenticate(self):
        self.assertFalse(self.service.verify(self.subject, '0'))

    def test_code_get_deleted_after_max_attempts(self):
        # The database entity must be deleted after MAX_ATTEMPTS
        # is reached, to prevent against enumeration attacks and guesses.
        self.assertTrue(LoginCode.objects.filter(subject=self.subject).exists())
        for i in range(LoginCode.MAX_ATTEMPTS + 1):
            self.assertFalse(self.service.verify(self.subject, '0'))
        self.assertFalse(LoginCode.objects.filter(subject=self.subject).exists())



# Generated by Django 3.0.4 on 2020-03-15 11:52

from django.db import migrations, models
import iam.domain.logincode
import unimatrix.lib.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('iam', '0006_principal'),
    ]

    operations = [
        migrations.CreateModel(
            name='LoginCode',
            fields=[
                ('subject', models.UUIDField(db_column='subject', primary_key=True, serialize=False)),
                ('code', models.CharField(db_column='code', default=iam.domain.logincode.LoginCode.new, max_length=127)),
                ('attempts', models.PositiveIntegerField(db_column='attempts', default=0)),
                ('generated', models.BigIntegerField(db_column='generated', default=unimatrix.lib.timezone.now)),
            ],
            options={
                'db_table': 'logincodes',
                'default_permissions': [],
            },
        ),
    ]



#pylint: skip-file
import django.test
import unimatrix.lib.test
from django.conf import settings

import iam.const
from iam.ext.django.models import Subject


@unimatrix.lib.test.integration
class RegistrationHappyFlowTestCase(django.test.TestCase):
    fixtures = ['testusers']

    def setUp(self):
        self.client = django.test.Client(HTTP_HOST='example.com')

    def test_register_with_email(self):
        response = self.client.post('/nl/registreren', {
            'principal': "happy.flow@test.unimatrixone.io",
            'password': 'foobarbaz',
            'first_name': 'Happy',
            'last_name': 'Flow'
        })

        # On succesful registration, the system must issue a redirect to
        # the default post-login page.
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['Location'], settings.LOGIN_REDIRECT_URL)

        # If the current request is authenticated, the login page must redirect
        # to the default postlogin page.
        response = self.client.get('/nl/aanmelden/')
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['Location'], settings.LOGIN_REDIRECT_URL)

        return response

    def test_can_login_after_registration(self):
        response = self.client.post('/nl/registreren', {
            'principal': "happy.flow@test.unimatrixone.io",
            'password': 'foobarbaz',
            'first_name': 'Happy',
            'last_name': 'Flow'
        })
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['Location'], settings.LOGIN_REDIRECT_URL)

        # Instantiate a new, unauthenticated client.
        client = django.test.Client(HTTP_HOST='example.com')
        response = client.get('/nl/aanmelden/')
        self.assertEqual(response.status_code, 200)

        response = client.post('/nl/aanmelden/', {
            'principal': "happy.flow@test.unimatrixone.io",
            'password': 'foobarbaz',
        })
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['Location'], settings.LOGIN_REDIRECT_URL)

    def test_register_with_email_and_next_parameter(self):
        response = self.client.post('/nl/registreren?next=/foo', {
            'principal': "happy.flow@test.unimatrixone.io",
            'password': 'foobarbaz',
            'first_name': 'Happy',
            'last_name': 'Flow'
        })

        # On succesful registration, the system must issue a redirect to
        # the default post-login page and redirect to the parameter provided
        # by ?next.
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['Location'], '/foo')

        # If the current request is authenticated, the login page must redirect
        # to the parameter specified by ?next.
        response = self.client.get('/nl/aanmelden/?next=/foo')
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['Location'], '/foo')

    def test_register_with_email_for_existing_user_redirects_to_default(self):
        # When a registration attempt is made with an email address that is
        # associated to an existing account, and the password matches, then
        # the application must redirect to the default login page.
        response = self.client.post('/nl/registreren', {
            'principal': iam.const.TEST_USER1_EMAIL,
            'password': iam.const.TEST_PASSWORD,
            'first_name': 'Happy',
            'last_name': 'Flow'
        })
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['Location'], settings.LOGIN_REDIRECT_URL)

    def test_register_with_email_for_existing_user_redirects_to_next(self):
        # When a registration attempt is made with an email address that is
        # associated to an existing account, and the password matches, then
        # the application must redirect to parameter specified by ?next.
        response = self.client.post('/nl/registreren?next=/foo', {
            'principal': iam.const.TEST_USER1_EMAIL,
            'password': iam.const.TEST_PASSWORD,
            'first_name': 'Happy',
            'last_name': 'Flow'
        })
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['Location'], '/foo')

    def test_register_with_email_for_existing_user_with_invalid(self):
        # When a registration attempt is made with an email address that is
        # associated to an existing account, and the password do not match,
        # then the application must not authenticate the user and display
        # a message that a confirmation email/sms is sent.
        response = self.client.post('/nl/registreren', {
            'principal': iam.const.TEST_USER1_EMAIL,
            'password': iam.const.TEST_PASSWORD + 'a',
            'first_name': 'Happy',
            'last_name': 'Flow'
        })
        self.assertEqual(response.status_code, 200)

    def test_register_does_not_create_subject_on_invalid_first_name(self):
        # When a registration attempt is made with an invalid form submission,
        # then a Subject instance must not be created.
        response = self.client.post('/nl/registreren', {
            'principal': "happy.flow@test.unimatrixone.io",
            'password': 'foobarbaz',
            'first_name': 'H',
            'last_name': 'Flow'
        })
        self.assertEqual(response.status_code, 200)
        self.assertTrue(not Subject.objects.filter(
            principals__value='happy.flow@test.unimatrixone.io').exists())

    def test_register_does_not_create_subject_on_invalid_last_name(self):
        # When a registration attempt is made with an invalid form submission,
        # then a Subject instance must not be created.
        response = self.client.post('/nl/registreren', {
            'principal': "happy.flow@test.unimatrixone.io",
            'password': 'foobarbaz',
            'first_name': 'Happy',
            'last_name': 'F'
        })
        self.assertEqual(response.status_code, 200)
        self.assertTrue(not Subject.objects.filter(
            principals__value='happy.flow@test.unimatrixone.io').exists())

    def test_register_does_not_create_subject_on_invalid_principal(self):
        # When a registration attempt is made with an invalid form submission,
        # then a Subject instance must not be created.
        response = self.client.post('/nl/registreren', {
            'principal': "123",
            'password': 'foobarbaz',
            'first_name': 'Happy',
            'last_name': 'F'
        })
        self.assertEqual(response.status_code, 200)
        self.assertTrue(not Subject.objects.filter(
            principals__value='123').exists())



"""Declares :class:`PermissionsMixin`."""
from django.contrib.auth.models import _user_get_permissions
from django.contrib.auth.models import _user_has_perm
from django.contrib.auth.models import _user_has_module_perms


class PermissionsMixin:

    def get_user_permissions(self, obj=None):
        """Return a list of permission strings that this user has directly.
        Query all available auth backends. If an object is passed in,
        return only permissions matching this object.
        """
        return _user_get_permissions(self, obj, 'user')

    def get_group_permissions(self, obj=None):
        """Return a list of permission strings that this user has through their
        groups. Query all available auth backends. If an object is passed in,
        return only permissions matching this object.
        """
        return _user_get_permissions(self, obj, 'group')

    def get_all_permissions(self, obj=None):
        return _user_get_permissions(self, obj, 'all')

    def has_perm(self, perm, obj=None):
        """Return True if the user has the specified permission. Query all
        available auth backends, but return immediately if any backend returns
        True. Thus, a user who has permission from a single auth backend is
        assumed to have permission in general. If an object is provided, check
        permissions for that object.
        """
        # Active superusers have all permissions.
        if self.is_active and self.is_superuser:
            return True

        # Otherwise we need to check the backends.
        return _user_has_perm(self, perm, obj)

    def has_perms(self, perm_list, obj=None):
        """Return True if the user has each of the specified permissions. If
        object is passed, check if the user has all required perms for it.
        """
        return all(self.has_perm(perm, obj) for perm in perm_list)

    def has_module_perms(self, app_label):
        """Return True if the user has any permissions in the given app label.
        Use similar logic as has_perm(), above.
        """
        # Active superusers have all permissions.
        if self.is_active and self.is_superuser:
            return True

        return _user_has_module_perms(self, app_label)
    



from django.apps import apps


class PrincipalAvailabilityService:

    def isavailable(self, realm, principal):
        """Return a bool indicating if the :term:`Principal` is available to
        add as an identifier for a :term:`Subject`.

        Args:
            principal (:class:`~iam.ext.domain.Principal`): the
                :term:`Principal` to identify a :term:`Subject` with.

        Returns:
            :obj:`bool`
        """
        Principal = apps.get_model('iam.Principal')
        return not Principal.objects.filter(
            realm_id=realm,
            kind=Principal.storage_kinds[principal.kind],
            value=principal.value
        ).exists()



"""Declares :class:`SubjectProperty`."""
from django.db import models


class SubjectProperty(models.Model):
    """Specifies the abstract interface for :term:`Subject Properties`."""

    id = models.BigAutoField(
        primary_key=True,
        blank=False,
        null=False,
        db_column='id'
    )

    subject = models.ForeignKey(
        'iam.Subject',
        blank=False,
        null=False,
        on_delete=models.CASCADE,
        db_column='subject_id'
    )

    kind = models.CharField(
        max_length=127,
        blank=False,
        null=False,
        db_column='kind'
    )

    class Meta:
        abstract = True



#
# Copyright (C) 2019-2020 Cochise Ruhulessin
#
# This file is part of python-iam.
#
# python-iam is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# python-iam is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with python-iam.  If not, see <https://www.gnu.org/licenses/>.
import uuid

from .resourcekind import ResourceKind
from .resourceuuid import ResourceUUID


class Resource:
    """A value-object representing a resource of a
    specific type.
    """

    @property
    def resource_class(self):
        return type(self).__name__

    @property
    def resource_id(self):
        return self.__resource_id

    @property
    def kind(self):
        return f'{self.api_version}/{self.resource_class}'

    @classmethod
    def fromdjango(cls, obj):
        """Create a :class:`Resource` instance from
        a Django model class or instance.
        """
        return cls(
            ResourceKind.fromdjango(obj),
            ResourceUUID.fromdjango(obj)
        )

    def __init__(self, resource_class, resource_id):
        self.__resource_class = resource_class
        self.__resource_id = resource_id

    def iscomposite(self):
        """Return a boolean indicating if the resource
        identity is a composity.
        """
        return False

    def __repr__(self):
        return f"<iam.{self.__class__.__name__}: {self.resource_id}>"



"""URL helper functions."""
import urllib.parse

import django.urls
from django.conf import settings


def updateqs(url, params):
    """Update the query parameters in the URL and return a string containing
    the new URL.
    """
    parts = list(urllib.parse.urlparse(url))
    qs = dict(urllib.parse.parse_qsl(parts[4]))
    qs.update(params)
    parts[4] = urllib.parse.urlencode(qs)
    return urllib.parse.urlunparse(parts)


def reverse(name, *args, **kwargs):
    return django.urls.reverse(f'{settings.IAM_URL_NAMESPACE}:{name}', *args, **kwargs)


def reverse_lazy(name, *args, **kwargs):
    return django.urls.reverse_lazy(
        f'{settings.IAM_URL_NAMESPACE}:{name}', *args, **kwargs)



"""Declares context processors for use with Django."""
from django.conf import settings
from django.utils.functional import SimpleLazyObject

from iam.ext.django.models import PlatformAgreementLanguage


def policies(request):
    """Updates the context with the effective :term:`Terms of Service` and
    :term:`Privacy Policy`.
    """
    f = lambda: PlatformAgreementLanguage.objects.get_effective(
        settings.IAM_TERMS_OF_SERVICE)
    p = lambda: PlatformAgreementLanguage.objects.get_effective(
        settings.IAM_PRIVACY_POLICY)
    return {
        'iam_terms_of_service': SimpleLazyObject(f),
        'iam_privacy_policy': SimpleLazyObject(p)
    }



"""Declares :class:`CredentialService`."""
import ioc

from iam.lib.meta import AttributeDispatcher


class CredentialService:
    """Provides an interface to verify and update credentials for specific
    :term:`Subjects`.
    """
    passwords = ioc.class_property('HashedPasswordRepository')
    otp = ioc.class_property('OneTimePasswordService')

    @AttributeDispatcher.dispatch('kind')
    def verify(self, func, credential, subject):
        return func(credential, subject)

    @verify.register('iam.unimatrixone.io/v1/PasswordCredential')
    def verify(self, credential, subject):
        secret = self.passwords.get(subject)
        return credential.check(secret.value)\
            if secret is not None\
            else False

    @verify.register('iam.unimatrixone.io/v1/LoginCodeCredential')
    def verify(self, credential, subject):
        return self.otp.verify(subject.pk, credential.secret)

    @verify.register('iam.unimatrixone.io/v1/NullCredential')
    def verify(self, credential, subject):
        return False



# Generated by Django 3.0.4 on 2020-03-18 18:02

from django.conf import settings
from django.db import migrations


def create_staff_realm(apps, schema_editor):
    Realm = apps.get_model('iam.Realm')
    Realm.objects.create(name=settings.IAM_STAFF_REALM, label='Backoffice')


class Migration(migrations.Migration):

    dependencies = [
        ('iam', '0012_subjectregistration'),
    ]

    operations = [
        migrations.RunPython(create_staff_realm)
    ]



#pylint: skip-file
import unittest

import unimatrix.lib.test

from ..emailprincipal import EmailPrincipal


@unimatrix.lib.test.unit
class EmailPrincipalDetectionTestCase(unittest.TestCase):
    """Tests that email addresses are detected by the ``iscapable()``
    method.
    """
    values = [
        'foo@bar.baz'
    ]
    cls = EmailPrincipal


    def test_iscapable(self):
        for value in self.values:
            self.assertTrue(self.cls.iscapable(value))



"""Declares :class:`JWTService`."""
import datetime
import hashlib
import json

import authlib.jose.errors
import jwt
import jwt.exceptions
from authlib.jose import JsonWebEncryption
from authlib.jose import JWE_ALGORITHMS
from django.conf import settings


class JWTService:
    default_ttl = 0

    @property
    def jwe(self):
        return JsonWebEncryption(algorithms=JWE_ALGORITHMS)

    @property
    def secret(self):
        return hashlib.md5(str.encode(settings.SECRET_KEY)).hexdigest()

    def encrypt(self, payload, sign=True, ttl=None):
        """Sign and encrypt `payload`."""
        protected = {'alg': 'A256GCMKW', 'enc': 'A256GCM'}
        if sign:
            payload = b'1' + self.sign(payload, ttl=ttl)
        else:
            payload = b'0' + str.encode(json.dumps(payload))
        return self.jwe.serialize_compact(protected, payload, self.secret)

    def decrypt(self, cipher):
        """Decrypts the JWE `payload` and verifies the enclosed JWS."""
        try:
            plain = self.jwe.deserialize_compact(cipher, self.secret)
            return plain['payload'][1:], bool(int(plain['payload'][0]))
        except authlib.jose.errors.DecodeError:
            return None, False

    def sign(self, payload, ttl=None):
        """Signs `payload`."""
        if ttl:
            payload['exp'] = datetime.datetime.utcnow()\
                + datetime.timedelta(seconds=ttl)
        return jwt.encode(payload, self.secret, algorithm='HS256')

    def decode(self, jws, verify=False):
        """Decodes a JSON Web Signature (JWS)."""
        return jwt.decode(jws, self.secret, algorithms=['HS256'])

    def verify(self, jws):
        """Verifies the payload."""
        if jws is None:
            return False
        try:
            self.decode(jws)
            return True
        except jwt.exceptions.PyJWTError:
            return False



"""Declares :class:`LoginCodeCredential`."""
from .credential import Credential


class LoginCodeCredential(Credential):

    @property
    def secret(self):
        return self.code

    def __init__(self, code):
        self.code = code



"""Declares :class:`BasePartialFormView`."""
from django.views.generic import FormView
from django.contrib.auth.decorators import login_required
from django.utils.decorators import method_decorator

from .iamcontextmixin import IAMContextMixin


@method_decorator(login_required, 'dispatch')
class BasePartialFormView(IAMContextMixin, FormView):
    """A :class:`django.forms.generic.FormView` implementation that renders
    only the form.
    """

    def form_invalid(self, *args, **kwargs):
        """Override the response code to return ``422``."""
        response = super().form_invalid(*args, **kwargs)
        response.status_code = 422
        return response

    def form_valid(self, form):
        """Return a HTTP response with the partial form."""
        return self.render_to_response(
            self.get_context_data(form=form))



# Generated by Django 3.0.4 on 2020-03-18 23:15
import datetime

from django.db import migrations, models
from django.utils import timezone


MIN_DATETIME = timezone.make_aware(datetime.datetime.min, timezone.utc)


def create_default_terms(apps, schema_editor):
    PlatformAgreement = apps.get_model('iam.PlatformAgreement')
    PlatformAgreement.objects.create(slug='terms',
        title="Terms of Service", effective=MIN_DATETIME)


def create_default_privacy_policy(apps, schema_editor):
    PlatformAgreement = apps.get_model('iam.PlatformAgreement')
    PlatformAgreement.objects.create(slug='privacy',
        title="Privacy Policy", effective=MIN_DATETIME)


class Migration(migrations.Migration):

    dependencies = [
        ('iam', '0016_auto_20200318_2146'),
    ]

    operations = [
        migrations.CreateModel(
            name='PlatformAgreement',
            fields=[
                ('slug', models.SlugField(db_column='slug', primary_key=True, serialize=False)),
                ('title', models.CharField(db_column='title', max_length=127)),
                ('effective', models.DateTimeField(db_column='effective')),
            ],
            options={
                'verbose_name': 'Platform Agreement',
                'verbose_name_plural': 'Platform Agreements',
                'db_table': 'platformagreements',
                'default_permissions': [],
            },
        ),
        migrations.RunPython(create_default_terms),
        migrations.RunPython(create_default_privacy_policy),
    ]



"""Declares authentication backends for the :mod:`iam.ext.django` Django
application.
"""
from iam.ext.django.models import Subject


class LocalSubjectBackend:
    model = Subject

    def authenticate(self, request, realm, principal, credential, **kwargs):
        """Authenticate a :term:`Subject` using the provided :term:`Principal`
        and :term:`Credential`.
        """
        try:
            subject = principal.resolve(realm, self.get_resolver())
        except self.model.DoesNotExist:
            credential.cooldown()
            return None
        else:
            if subject.check_credential(credential)\
            and self.can_authenticate(subject):
                return subject

    def can_authenticate(self, subject):
        """Return a boolean indicating if a :term:`Subject` may
        authenticate.
        """
        return True

    def get_user(self, subject_id):
        """Return a :class:`Subject` instance using the given `subject_id`."""
        return self.model.objects.get(pk=subject_id)

    def get_resolver(self):
        """Return an instance that can resolve a :class:`~iam.domain.Principal`."""
        return self.model.objects



"""Declares form fields and base classes."""
from django import forms
from django.core.validators import MinLengthValidator
from django.utils.translation import gettext as _


def notallupper(value):
    """Ensure that `value` does not consist of only uppercase characters."""
    if str.upper(value) == value:
        raise forms.ValidationError(_("Do not use only uppercase characters."))


def notalllower(value):
    """Ensure that `value` does not consist of only lowercase characters."""
    if str.lower(value) == value:
        raise forms.ValidationError(_("Do not use only lowercase characters."))


def excludeforbidden(value):
    """Ensure that `value` does not contain forbidden characters."""
    forbidden = set([';','<','>'])
    if (set(value) & forbidden):
        raise forms.ValidationError("Stop probing.")


class NameField(forms.CharField):
    """A :class:`~django.forms.CharField` implementation that is used for
    first and last names.
    """

    def __init__(self, *args, **kwargs):
        validators = kwargs.setdefault('validators', [])
        validators.extend([
            MinLengthValidator(2),
            notallupper,
            notalllower,
            excludeforbidden
        ])
        super().__init__(*args, **kwargs)



#pylint: skip-file
import unittest

import unimatrix.lib.test
from django.conf import settings
from django.test import RequestFactory

from iam.const import MESSAGE_QUERY_PARAM
from ..auth import on_redirected


@unimatrix.lib.test.unit
class RedirectMessageVerificationTestCase(unittest.TestCase):
    """Tests if a message is correctly verified on SSO redirect."""

    def setUp(self):
        self.factory = RequestFactory()

    def test_no_message_is_ignored(self):
        request = self.factory.get('/')
        self.assertIsNone(on_redirected(request))

    def test_invalid_message_interrupts_pipeline(self):
        request = self.factory.get('/', {MESSAGE_QUERY_PARAM: 'ffoo'})



#pylint: skip-file
import django.test
import unimatrix.lib.test
from django.contrib.auth import authenticate
from django.contrib.auth import login

from iam.const import TEST_USER1_EMAIL
from iam.const import TEST_USER1_ID
from iam.const import TEST_PASSWORD
from iam.domain import EmailPrincipal
from iam.domain import NullCredential
from iam.domain import PasswordCredential
from iam.ext.django.lib.backends import LocalSubjectBackend


@unimatrix.lib.test.integration
class PasswordAuthenticationTestCase(django.test.TestCase):
    fixtures = ['testusers']
    authenticate_kwargs = {
        'request': None,
        'principal': EmailPrincipal(TEST_USER1_EMAIL),
        'credential': PasswordCredential(TEST_PASSWORD)
    }

    def setUp(self):
        self.backend = LocalSubjectBackend()

    def get_authenticate_kwargs(self):
        return dict(self.authenticate_kwargs)

    def test_login_with_valid_credentials(self):
        sub = self.backend.authenticate(**self.get_authenticate_kwargs())
        self.assertIsNotNone(sub)
        self.assertEqual(str(sub.pk), TEST_USER1_ID)

    def test_login_with_invalid_credentials(self):
        params = self.get_authenticate_kwargs()
        params['credential'] = PasswordCredential('wrong password')
        sub = self.backend.authenticate(**params)
        self.assertIsNone(sub)

    def test_login_with_no_credentials(self):
        params = self.get_authenticate_kwargs()
        params['credential'] = PasswordCredential('')
        sub = self.backend.authenticate(**params)
        self.assertIsNone(sub)

    def test_login_with_none_credentials(self):
        params = self.get_authenticate_kwargs()
        params['credential'] = None
        sub = self.backend.authenticate(**params)
        self.assertIsNone(sub)

    def test_login_with_null_credentials(self):
        params = self.get_authenticate_kwargs()
        params['credential'] = NullCredential()
        sub = self.backend.authenticate(**params)
        self.assertIsNone(sub)



"""Declares :class:`~iam.ext.django.models.EmailPrincipal`."""
from django.db import models


class EmailPrincipal(models.Model):
    """Maintains an email address and associated properties for a :term:`Subject`
    within a :term:`Realm`.
    """

    #: Specifies the :class:`~iam.ext.django.models.Realm` to which the
    #: :term:`Email Principal` belongs.
    realm = models.ForeignKey(
        'iam.Realm',
        on_delete=models.PROTECT,
        blank=False,
        null=False,
        db_column='realm_name'
    )

    #: Specifies the :term:`Subject` that is identified by the email address.
    subject = models.ForeignKey(
        'iam.Subject',
        on_delete=models.CASCADE,
        blank=False,
        null=False,
        related_name='email',
        db_column='subject_id'
    )

    value = models.EmailField(
        blank=False,
        null=False,
        db_column='value'
    )

    def save(self, *args, **kwargs):
        if self._state.adding:
            self.realm_id = self.subject.realm_id
        return super().save(*args, **kwargs)

    class Meta:
        app_label = 'iam'
        db_table = 'emailprincipals'
        default_permissions = []

        #: TODO: Worst normalization ever. A both realms and subjects can not
        #: have duplicate emails, but no composite keys in Django FTW.
        unique_together = [
            ('realm', 'value'),
            ('subject', 'value')
        ]





class AbstractRealm:
    """Abstract implementation of an authorization realm."""

    def get_principal_roles(self, principal):
        """Return an iterable yielding the roles that were
        assigned to the principal in this realm.
        """
        raise NotImplementedError



from .agreementacceptance import AgreementAcceptanceService
from .credential import CredentialService
from .jwt import JWTService
from .onetimepassword import OneTimePasswordService
from .password import PasswordService
from .passwordreset import PasswordResetService
from .principalavailability import PrincipalAvailabilityService
from .registration import RegistrationService
from .template import TemplateService



"""Declares :class:`PasswordCredential`."""
from django.contrib.auth.hashers import check_password
from django.contrib.auth.hashers import make_password

from .credential import Credential


class PasswordCredential(Credential):

    @property
    def secret(self):
        """Return the secret associated to this :term:`Credential`."""
        return make_password(self.raw_password)

    def __init__(self, raw_password):
        self.raw_password = raw_password
        self.new_password = None

    def cooldown(self):
        """Invoked when the :term:`Principal` with which an authentication
        attempt is being made does not resolve to a user. This is to prevent
        against timing attacks et. al.
        """
        make_password(self.raw_password)

    def check(self, password):
        """Verifies that :attr:`raw_password` matches `password`."""
        if password is None:
            return None

        return check_password(self.raw_password, password,
            lambda x: setattr(self, 'new_password', make_password(x)))



"""Declares :class:`AgreementAcceptanceService`."""
from django.apps import apps


class AgreementAcceptanceService:
    """Provides an interface to accept :term:`Platform Agreements`."""

    def accept(self, checksum, subject, host):
        """Accepts the :term:`Platform Agreement` `oid` for a
        :term:`Subject`.

        Args:
            checksum (:obj:`int`): the checksum of the agreement to accept.
            subject (:obj:`uuid.UUID`): identifies the :term:`Subject`.
            host (:obj:`str`): the remote host address (IPv4/IPv6) from
                which the HTTP request to accept the agreement was made.

        Returns:
            :class:`types.NoneType`
        """
        AcceptedPlatformAgreement = apps.get_model(
            'iam.AcceptedPlatformAgreement')

        AcceptedPlatformAgreement.objects.create(agreement_id=checksum,
            subject_id=subject, host=host)



"""Declares :class:`LoginView`."""
from django import forms
from django.conf import settings
from django.contrib.auth import logout
from django.contrib.auth.views import LoginView as DjangoLoginView
from django.utils.translation import gettext_lazy as _

from .forms import ChoosePasswordForm
from .forms import PolymorphicPrincipalForm
from .subjectonboarding import SubjectOnboardingMixin


class LoginView(SubjectOnboardingMixin, DjangoLoginView):
    """Exposes ``GET`` and ``POST`` methods to login a :term:`Subject`."""
    template_name = 'iam/login.html.j2'
    form_class = PolymorphicPrincipalForm
    allow_signup = True

    # TODO: Implement a mechanism to keep the below methods in sync.
    def must_redirect(self):
        """Return a boolean indicating if we must redirect an authenticated user
        away from this page.
        """
        subject = self.request.user
        redirect = super().must_redirect()
        if not subject.is_authenticated:
            return redirect

        # If the Subject has not accepted the terms of service, then it
        # should be prompted to do so. Same applies to the privacy policy.
        if not subject.accepts(settings.IAM_TERMS_OF_SERVICE)\
        or not subject.accepts(settings.IAM_PRIVACY_POLICY):
            return False

        # Check if the user belongs to the realm specified by this
        # authentication gateway. If the user authenticated in a different,
        # log it out and provide an opportunity to login at this realm.
        if subject.realm_id != self.realm:
            logout(self.request)
            return False


        # If the user logged in with SSO then it should not be prompted to
        # set a password.
        if not subject.haspassword() and not self.request.sso:
            redirect = False

        # If the user has no first name or last name, prompt it to set it.
        if not subject.first_name or not subject.last_name:
            redirect = False

        return redirect

    def get_form_class(self):
        """Return the appropriate form based on the ``op`` parameter in the
        POST-data.
        """
        form = super().get_form_class()
        if self.request.user.is_authenticated\
        and not self.request.user.haspassword()\
        and not self.request.sso:
            form = ChoosePasswordForm
            self.must_login = False

        form.allow_create = self.allow_signup
        return form

    def get_template_names(self):
        """Return a list of template names."""
        template_names = []
        if self.request.user.is_authenticated:
            subject = self.request.user
            if not subject.haspassword()\
            and not self.request.sso:
                # The user must choose a new password.
                template_names = ['iam/login.passwordexpired.html.j2']
            if not subject.accepts(settings.IAM_TERMS_OF_SERVICE):
                template_names = 'iam/login.useragreement.html.j2'
            elif not subject.accepts(settings.IAM_PRIVACY_POLICY):
                template_names = 'iam/login.useragreement.privacy.html.j2'
            elif not subject.first_name or not subject.last_name:
                template_names = ['iam/login.completenames.html.j2']
        else:
            template_names = super().get_template_names()
        return template_names



# Generated by Django 3.0.4 on 2020-03-18 10:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('iam', '0010_subject_kind'),
    ]

    operations = [
        migrations.AlterField(
            model_name='principal',
            name='id',
            field=models.BigAutoField(db_column='id', primary_key=True, serialize=False),
        ),
    ]



from .acceptagreement import AcceptAgreementFormView
from .choosenames import ChooseNamesFormView
from .debug import DebugView
from .login import LoginView
from .logintoken import LoginTokenView
from .magiclink import MagicLinkView
from .register import RegisterView
from .resetcredentials import ResetCredentialsView



"""Declares programming helper functions."""
import functools
import types


class AttributeDispatcher:
    """Decorates a dispatching method on a class and provides an interface
    to register handler functions.
    """

    @classmethod
    def dispatch(cls, attname):
        """Decorates the dispatching method to inject `func` based on the
        internal mapping lookup using the attribute or method `attname` of
        the pre-decoration first positional argument to the decorated
        method.
        """
        def decorator_factory(func):
            return functools.wraps(func)(cls(attname, func))
        return decorator_factory

    def __init__(self, attname, dispatcher):
        self.dispatcher = dispatcher
        self.attname = attname
        self.handlers = {}

    def get_handler(self, obj):
        """Returns the appropriate handler based on the value of attribute
        `attname` on `obj`.
        """
        name = getattr(obj, self.attname)
        if callable(name):
            name = name()
        return self.handlers[name]

    def register(self, name):
        """Registers a handler to the dispatcher."""
        def decorator_factory(decoratable):
            self.handlers[name] = decoratable

            @functools.wraps(decoratable)
            def f(this, obj, *args, **kwargs):
                return self.get_handler(obj)(this, obj, *args, **kwargs)
            f.register = self.register
            return f
        return decorator_factory



from .acceptedplatformagreements import AcceptedPlatformAgreement
from .passwordcredentials import PasswordCredential
from .emailprincipals import EmailPrincipal
from .logincodes import LoginCode
from .platformagreements import PlatformAgreement
from .platformagreementlanguages import PlatformAgreementLanguage
from .principals import Principal
from .realms import Realm
from .subjects import Subject
from .subjectregistrations import SubjectRegistration
from .subjectbooleanproperties import SubjectBooleanProperty
from .subjectstringproperties import SubjectStringProperty



"""Declares view decorators."""
import functools
import urllib.parse

from django.conf import settings
from django.http import HttpResponseRedirect


def requireconsent(agreement):
    """Requires the :term:`Subject` to agree with the given agreement."""
    def decorator(func):
        @functools.wraps(func)
        def f(request, *args, **kwargs):
            redirect_to = settings.LOGIN_URL\
                + '?'\
                + urllib.parse.urlencode({'next': request.get_full_path()})
            if request.user.is_authenticated\
            and not request.user.accepts(agreement):
                return HttpResponseRedirect(redirect_to)
            return func(request, *args, **kwargs)
        return f
    return decorator



"""Declares :class:`RegisterView`."""
from django import forms
from django.core.validators import MinLengthValidator

from ..lib.forms import NameField
from .forms import PolymorphicPrincipalForm
from .login import LoginView
from .subjectonboarding import SubjectOnboardingMixin


class RegisterView(LoginView):
    """Exposes ``GET`` and ``POST`` methods to register a new
    :term:`Subject`.
    """
    template_name = 'iam/register.html.j2'

    class form_class(PolymorphicPrincipalForm):
        require_valid_credentials = False
        allow_create = True

        first_name = NameField(
            max_length=32,
            required=True,
        )

        last_name = NameField(
            max_length=32,
            required=True,
        )

        def create_subject(self, *args, **kwargs):
            """Creates a new :term:`Subject` that is identified by the given
            :term:`Principal` `principal` and :term:`Password` `password`. Return
            the :class:`~iam.ext.django.models.Subject` instance, or ``None``
            when the :term:`Principal` is already associated to a :term:`Subject`.
            """
            subject = super().create_subject(*args, **kwargs)
            if subject:
                subject.first_name = self.cleaned_data['first_name']
                subject.last_name = self.cleaned_data['last_name']
            return subject

    def form_valid(self, form):
        """All security checks complete; log in the user."""
        assert not self.request.user.is_authenticated,\
            "Subject should not be authenticated when using this form."
        subject = form.get_user()
        return super().form_valid(form)\
            if not subject.is_anonymous\
            else super().form_invalid(form)





class Authorizable:
    """Represents an object that is authorizable."""

    @classmethod
    def get_resource_class(cls):
        """Returns the resource class of the objec."""
        raise NotImplementedError

    def get_resource_qualname(self):
        """Returns the qualified name of a resource,
        identifying it within the global scope of
        a system.
        """
        raise NotImplementedError

    def get_subject_authorizations(self, subject):
        """Return a set holding the authorizations that
        the subject has on this entity. The default
        implementation returns an empty set, meaning
        no permissions.
        """
        realm = self.get_realm()
        return (
            self.get_resource_qualname(),
            realm.get_realm_authorizations(subject, self)\
            if realm else set()
        )

    def get_realm(self):
        """Returns the authorization realm to which this
        resource belongs.
        """
        raise NotImplementedError



# Generated by Django 3.0.4 on 2020-03-17 23:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('iam', '0009_auto_20200316_0256'),
    ]

    operations = [
        migrations.AddField(
            model_name='subject',
            name='kind',
            field=models.CharField(db_column='kind', default='User', max_length=63),
        ),
    ]



"""Declares the :class:`Realm` model."""
from django.db import models
from django.conf import settings

from iam.canon.dto import RealmConfigSchema



class Realm(models.Model):
    """A :class:`Realm` manages a set of users, credentials, roles, and groups.
    A :term:`Subject` belongs to and logs into a :term:`Realm`. Realms are
    isolated from one another and can only manage and authenticate the users
    that they control.
    """

    @property
    def config(self):
        schema = RealmConfigSchema()
        return schema.load(settings.IAM_REALM_SETTINGS.get(self.name) or {})

    #: A symbolic name for the realm. The :attr:`name` attribute may consist
    #: of alphanumeric characters and start with a letter. It can not be changed
    #: after creating the realm.
    name = models.CharField(
        max_length=64,
        primary_key=True,
        blank=False,
        null=False,
        db_column='name'
    )

    #: A human-readable name for the realm.
    label = models.CharField(
        max_length=64,
        blank=False,
        null=False,
        db_column='label'
    )

    def allowsso(self, provider):
        """Return a boolean indicating if this :term:`Realm` allow
        authentication using the specified SSO `provider`.
        """
        return provider in self.config.sso.providers

    def getssoproviders(self):
        """Return a tuple containg all SSO providers for this realm."""
        return tuple(self.config.sso.providers)

    def __str__(self):
        return self.name

    class Meta:
        app_label = 'iam'
        db_table = 'realms'
        default_permissions = []



# Generated by Django 3.0.4 on 2020-03-16 02:52

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('iam', '0007_logincode'),
    ]

    operations = [
        migrations.CreateModel(
            name='SubjectStringProperty',
            fields=[
                ('id', models.BigAutoField(db_column='id', primary_key=True, serialize=False)),
                ('kind', models.CharField(db_column='kind', max_length=127)),
                ('value', models.CharField(db_column='value', max_length=127)),
                ('subject', models.ForeignKey(db_column='subject_id', on_delete=django.db.models.deletion.CASCADE, related_name='stringproperties', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'subjectstringproperties',
                'default_permissions': [],
                'default_related_name': 'stringproperties',
            },
        ),
    ]



"""Declares :class:`PasswordService`."""
import ioc
from django.apps import apps

from iam.domain import PasswordCredential


class PasswordService:
    """Exposes an interface to change and expire passwords."""
    repo = ioc.class_property('SubjectRepository')

    def change(self, subject, password):
        """Change the password for :term:`Subject` `subject`.

        Args:
            subject (:class:`uuid.UUID`): identifies the :term:`Subject` to
                generate a code for.
            password (:obj:`str`): the new password to set for the
                :term:`Subject`.

        Returns:
            :class:`types.NoneType`
        """
        subject = self.repo.get(subject)
        subject.addcredential(PasswordCredential(password))



# Generated by Django 3.0.4 on 2020-03-07 00:21

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Realm',
            fields=[
                ('name', models.CharField(db_column='name', max_length=64, primary_key=True, serialize=False)),
                ('label', models.CharField(db_column='label', max_length=64)),
            ],
            options={
                'db_table': 'realms',
                'default_permissions': [],
            },
        ),
        migrations.CreateModel(
            name='Subject',
            fields=[
                ('id', models.UUIDField(db_column='id', default=uuid.uuid4, primary_key=True, serialize=False)),
                ('realm', models.ForeignKey(db_column='realm_name', default='default', on_delete=django.db.models.deletion.PROTECT, to='iam.Realm')),
            ],
            options={
                'db_table': 'subjects',
                'default_permissions': [],
            },
        ),
    ]



"""Declares :class:`SubjectOnboardingMixin`."""
from django.http import HttpResponseRedirect
from django.utils.translation import gettext_lazy as _
from django.db import transaction
from django.conf import settings
from django.contrib.auth.views import LoginView
from django.contrib.auth.views import SuccessURLAllowedHostsMixin
from django.shortcuts import resolve_url
from django.utils.decorators import method_decorator
try:
    from django.utils.http import url_has_allowed_host_and_scheme
except ImportError:
    # Not available before Django 3.0
    url_has_allowed_host_and_scheme = lambda *a, **k: True
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_protect
from django.views.decorators.debug import sensitive_post_parameters
from django.views.generic import FormView

from .iamcontextmixin import IAMContextMixin


class SubjectOnboardingMixin(IAMContextMixin, SuccessURLAllowedHostsMixin):
    """Mixin class that provides extra context to render login and registration
    pages.
    """
    must_login = True
    redirect_field_name = LoginView.redirect_field_name
    redirect_authenticated_user = settings.IAM_REDIRECT_AUTHENTICATED_USER

    @method_decorator(transaction.atomic)
    @method_decorator(sensitive_post_parameters())
    @method_decorator(csrf_protect)
    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        if self.must_redirect():
            redirect_to = self.get_success_url()
            if redirect_to == self.request.path:
                raise ValueError(
                    "Redirection loop for authenticated user detected. Check that "
                    "your LOGIN_REDIRECT_URL doesn't point to a login page."
                )
            return HttpResponseRedirect(redirect_to)
        return FormView.dispatch(self, request, *args, **kwargs)

    def must_redirect(self):
        """Return a boolean indicating if we must redirect an authenticated user
        away from this page.
        """
        return self.redirect_authenticated_user\
            and self.request.user.is_authenticated

    def form_valid(self, form):
        """All security checks complete; log in the user."""
        if self.must_login:
            response = super().form_valid(form)
        else:
            response = HttpResponseRedirect(self.get_success_url())

        # Set appropriate status code to prevent browsers from caching the
        # redirect URI, if any.
        response.status = 303
        return response

    def get_form_kwargs(self):
        """Return the keyword arguments used to instantiate the form
        class.
        """
        kwargs = super().get_form_kwargs()
        kwargs['request'] = self.request
        kwargs['realm'] = self.get_realm()
        return kwargs

    def get_success_url(self):
        """Return the success URL to redirect when a valid form is
        submitted.
        """
        url = self.get_redirect_url()
        return url or resolve_url(settings.LOGIN_REDIRECT_URL)

    def get_redirect_url(self):
        """Return the user-originating redirect URL if it's safe."""
        redirect_to = self.request.POST.get(
            self.redirect_field_name,
            self.request.GET.get(self.redirect_field_name, '')
        )
        url_is_safe = url_has_allowed_host_and_scheme(
            url=redirect_to,
            allowed_hosts=self.get_success_url_allowed_hosts(),
            require_https=self.request.is_secure(),
        )
        return redirect_to if url_is_safe else ''

    def get_template_names(self):
        """Return a list of template names. The default implementation simply
        returns :attr:`template_name`.
        """
        return [self.template_name]



# Generated by Django 3.0.4 on 2020-03-16 02:56

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('iam', '0008_subjectstringproperty'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='subjectstringproperty',
            unique_together={('subject', 'kind')},
        ),
    ]



"""Declares class:`PasswordResetService`."""
import ioc
from django.core.mail import send_mail
from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import gettext as _

import iam.lib.urls
from iam.lib.meta import AttributeDispatcher


class PasswordResetService:
    """Exposes an interface to reset passwords."""
    otp = ioc.class_property('OneTimePasswordService')
    template = ioc.class_property('iam.TemplateService')

    @AttributeDispatcher.dispatch('kind')
    def challenge(self, func, principal):
        return func(principal)

    @challenge.register('iam.unimatrixone.io/v1/EmailPrincipal')
    def challenge(self, principal, verify=None, context_data=None):
        """Challenge the :term:`Subject` by sending a :term:`One Time Password`
        to the :term:`Principal`. Return a boolean indicating if a challenge
        was sent.

        Args:
            principal (:class:`~iam.domain.Principal`): the :term:`Principal`
                to send the :term:`One Time Password` to. It must be a valid
                :term:`Contact Mechanism`.
            verify (:obj:`str`): a string holding an URL that the
                :term:`Subject` may visit to continue the password reset
                process.

        Returns:
            :obj:`bool`
        """
        try:
            principal.subject.password.delete() # TODO: move to PasswordService
        except (AttributeError, ObjectDoesNotExist):
            pass
        code = self.otp.generate(principal.subject.pk)
        ctx = context_data or {}

        # Append the code to the URI so that the endpoint can detect is.
        if verify:
            verify = iam.lib.urls.updateqs(verify, {
                'token': code,
                'principal': principal.value
            })

        ctx.update({
            'principal': principal,
            'subject': principal.subject,
            'code': code,
            'url': verify
        })

        principal.send(
            subject=_("Reset your password"),
            text=self.template.render_to_string('email/resetpassword.body.txt.j2', ctx))

    def verify(self, subject, code):
        """Verify that the given `code` matches the one sent in the challenge.
        If `code` is valid, mark the :term:`Password` associated to the
        :term:`Subject` as expired, so it will be forced to choose a new one,
        and return a boolean indicating if the code verification was a succes.

        Args:
            subject (:class:`uuid.UUID`): a Universally Unique Identifier (UUID)
                identifying the :term:`Subject` that was challenged.
            code (:obj:`str`): the secret that was sent to the :term:`Subject`.

        Returns:
            :obj:`bool`
        """
        return False



"""Declares :class:`MagicLinkView`."""
from django import forms
from django.views.generic import FormView

from .login import LoginView


class MagicLinkView(LoginView):
    """Provides a mechanism to send a magic link for login purposes to
    known users.
    """
    template_name = 'iam/login.magic.html.j2'

    class form_class(forms.Form):
        email = forms.EmailField(
            required=True
        )



"""Declares :class:`LoginCode`."""
import secrets


class LoginCode:
    """Represents a code that a :term:`Subject` may use to authenticate itself
    in a temporary fashion.
    """

    @classmethod
    def new(cls):
        """Return a new :class:`LoginCode` with a random value."""
        return cls(str(secrets.randbelow(999999)).zfill(6))

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return self.value



"""Declares :class:`NullCredential`."""
from .credential import Credential


class NullCredential(Credential):
    """A :class:`~iam.domain.Credential` implementation that never
    verifies.
    """

    def __init__(self, *args, **kwargs):
        pass

    def check(self, *args, **kwargs):
        """Verifies the credential. Always returns ``False``."""
        return False



"""Declares :class:`Credential`."""
from .resource import Resource


class Credential(Resource):
    """The base class for all :term:`Credentials`."""
    api_version = 'iam.unimatrixone.io/v1'

    def cooldown(self):
        """Invoked when the :term:`Principal` with which an authentication
        attempt is being made does not resolve to a user. This is to prevent
        against timing attacks et. al.
        """
        return



"""Declares :class:`ExternalIdentifierPrincipal`."""
from .principal import Principal


class ExternalIdentifierPrincipal(Principal):
    """Represents an identifier that is external."""
    platform = None

    def inrealm(self, realm):
        """Return a string holding an identifier within a realm."""
        assert self.platform is not None
        return f'realm+sso://{self.platform}@{realm}/{self.value}'



"""Declares :class:`PasswordCredential`."""
from django.db import models


class PasswordCredential(models.Model):
    """Maintains hashed passwords and basic properties, such as an expiration
    date.
    """
    subject = models.OneToOneField(
        'iam.Subject',
        on_delete=models.CASCADE,
        primary_key=True,
        blank=False,
        null=False,
        related_name='password',
        db_column='subject_id'
    )

    value = models.CharField(
        max_length=128,
        blank=False,
        null=False,
        db_column='value'
    )

    class Meta:
        app_label = 'iam'
        db_table = 'passwordcredentials'
        default_permissions = []
